import os
import pathlib
from typing import Union

import f90nml  # type: ignore[import]

from pyschism.enums import NWSType
from pyschism.mesh.base import Gr3
from pyschism.forcing.atmosphere.nws.nws import NWS
from pyschism.forcing.atmosphere.nws.nws2.sflux import SfluxDataset


SFLUX_DEFAULTS = f90nml.read(
    pathlib.Path(__file__).parent / 'sflux_inputs.txt')


class Sflux1Descriptor:

    def __set__(self, obj, sflux_1: SfluxDataset):
        if not isinstance(sflux_1, SfluxDataset):
            raise TypeError(f'Argument sflux_1 must be of type {SfluxDataset},'
                            f' not type {type(sflux_1)}')
        obj.__dict__['sflux_1'] = sflux_1

    def __get__(self, obj, val):
        return obj.__dict__['sflux_1']


class Sflux2Descriptor:

    def __set__(self, obj, sflux_2: Union[SfluxDataset, None]):
        if sflux_2 is not None:
            if not isinstance(sflux_2, SfluxDataset):
                raise TypeError(f'Argument sflux_2 must be of type {SfluxDataset},'
                                f' not type {type(sflux_2)}.')
            obj.__dict__['sflux_2'] = sflux_2

    def __get__(self, obj, val):
        return obj.__dict__.get('sflux_2')


class WindRotDescriptor:

    def __set__(self, obj, grd: Gr3):
        if not isinstance(grd, Gr3):
            raise TypeError(f'Argument grd must be of type {Gr3}, '
                            f'not type {type(grd)}.')
        data = grd.to_dict()
        data.pop("boundaries", None)
        data["nodes"] = {id: (coord, 0.) for id, (coord, _)
                         in data["nodes"].items()}
        obj.__dict__['windrot'] = Gr3(**data)

    def __get__(self, obj, val):
        return obj.__dict__.get('windrot')


class NWS2(NWS):

    _sflux_1 = Sflux1Descriptor()
    _sflux_2 = Sflux2Descriptor()
    _windrot = WindRotDescriptor()

    def __init__(self, sflux_1: SfluxDataset,
                 sflux_2: SfluxDataset = None):
        """Loads SfluxDataset to use as NWS2 input. """

        # for key, item in SFLUX_DEFAULTS['sflux_inputs'].items():
        #     print(key, item)
        #     setattr(self, key, item)

        self._sflux_1 = sflux_1
        self._sflux_2 = sflux_2

    def __call__(self, model_driver):
        super().__call__(model_driver)
        if hasattr(self.sflux_1, 'fetch_data'):
            self.sflux_1.fetch_data(
                self._start_date, self._rnday,
                bbox=model_driver.model_domain.hgrid.get_bbox('EPSG:4326'))
        if self.sflux_2 is not None:
            if hasattr(self.sflux_2, 'fetch_data'):
                self.sflux_2.fetch_data(
                    self._start_date, self._rnday,
                    bbox=model_driver.model_domain.hgrid.get_bbox('EPSG:4326'))
        self._windrot = model_driver.model_domain.hgrid

    def __str__(self):
        data = []
        data = '\n'.join(data)
        return f'&sflux_inputs\n{data}/\n'

    def write(self, path: Union[str, os.PathLike], overwrite: bool = False,
              wind_rot: bool = True):
        # write sflux namelist
        path = pathlib.Path(path)
        if path.name != 'sflux':
            path /= 'sflux'
        path.mkdir(exist_ok=True)
        with open(path / 'sflux_inputs.txt', 'w') as f:
            f.write(str(self))
        # write sflux data
        self.sflux_1.write(
            path, 1, overwrite, start_date=self._start_date, rnday=self._rnday)
        if self.sflux_2 is not None:
            self.sflux_2.write(
                path, 2, overwrite, start_date=self._start_date,
                rnday=self._rnday)
        # # write windrot data
        if self._windrot is not None:
            if wind_rot is True:
                self._windrot.write(path.parent / 'windrot_geo2proj.gr3',
                                    overwrite)

    @property
    def dtype(self) -> NWSType:
        """Returns the datatype of the object"""
        return NWSType(2)

    @property
    def sflux_1(self) -> SfluxDataset:
        return self._sflux_1

    @property
    def sflux_2(self) -> SfluxDataset:
        return self._sflux_2



#! /usr/bin/env python
import tempfile
import pathlib
from pyschism.mesh.friction import Fgrid
from pyschism.mesh import Mesh, Vgrid
from unittest.mock import patch
import unittest


class MeshTestCase(unittest.TestCase):

    def setUp(self):
        nodes = {
            '1': ((0., 0.), 1),
            '2': ((.5, 0.), -4.),
            '3': ((1., 0.), -3.),
            '4': ((1., 1.), -8.),
            '5': ((0., 1.), 1.),
            '6': ((.5, 1.5), -9.),
            '7': ((.33, .33), 1.),
            '8': ((.66, .33), 2.),
            '9': ((.5, .66), 3.),
            '10': ((-1., 1.), 4.),
            '11': ((-1., 0.), 5.),
            }
        elements = {
            '1': ['5', '7', '9'],
            '2': ['1', '2', '7'],
            '3': ['2', '3', '8'],
            '4': ['8', '7', '2'],
            '5': ['3', '4', '8'],
            '6': ['4', '9', '8'],
            '7': ['4', '6', '5'],
            '8': ['5', '10', '11', '1'],
            '9': ['9', '4', '5'],
            '10': ['5', '1', '7']
            }
        # write hgrid
        tmpdir = tempfile.TemporaryDirectory()
        hgrid = pathlib.Path(tmpdir.name) / 'hgrid.gr3'
        with open(hgrid, 'w') as f:
            f.write('\n')
            f.write(f'{len(elements):d} ')
            f.write(f'{len(nodes):d}\n')
            for id, ((x, y), z) in nodes.items():
                f.write(f"{id} ")
                f.write(f"{x} ")
                f.write(f"{y} ")
                f.write(f"{z}\n")
            for id, geom in elements.items():
                f.write(f"{id} ")
                f.write(f"{len(geom)} ")
                for idx in geom:
                    f.write(f"{idx} ")
                f.write(f"\n")

        vgrid = pathlib.Path(tmpdir.name) / 'vgrid.gr3'
        with open(vgrid, 'w') as f:
            f.write("2 !ivcor\n")
            f.write("2 1 1.e6 !nvrt, kz (# of Z-levels); h_s (transition depth between S and Z)\n")
            f.write("Z levels\n")
            f.write("1  -1.e6\n")
            f.write("S levels\n")
            f.write("40. 1. 1.e-4  !h_c, theta_b, theta_f\n")
            f.write("   1    -1.\n")
            f.write("   2     0.\n")
        fgrid = pathlib.Path(tmpdir.name) / 'fgrid.gr3'
        with open(fgrid, 'w') as f:
            f.write('generic_friction\n')
            f.write(f'{len(elements):d} ')
            f.write(f'{len(nodes):d}\n')
            for id, ((x, y), z) in nodes.items():
                f.write(f"{id} ")
                f.write(f"{x} ")
                f.write(f"{y} ")
                f.write(f"{z}\n")
            for id, geom in elements.items():
                f.write(f"{id} ")
                f.write(f"{len(geom)} ")
                for idx in geom:
                    f.write(f"{idx} ")
                f.write(f"\n")
        self.tmpdir = tmpdir
        self.hgrid = hgrid
        self.vgrid = vgrid
        self.fgrid = fgrid

    def test_open(self):
        self.assertIsInstance(
            Mesh.open(
                self.hgrid,
                self.vgrid,
                self.fgrid
                ),
            Mesh
        )

    def test_make_plot(self):
        m = Mesh.open(self.hgrid)
        m.make_plot()

    def test_default_fgrid(self):
        m = Mesh.open(self.hgrid)
        assert isinstance(m.fgrid, Fgrid)

    @patch.object(Vgrid, 'is3D', return_value=True)
    def test_make_plot_3D(self, mock):
        m = Mesh.open(self.hgrid)
        self.assertRaises(NotImplementedError, m.make_plot)

    def test__fgrid_getter(self):
        h = Mesh.open(self.hgrid)
        h.fgrid
        self.assertIsInstance(h._fgrid, Fgrid)

    def test_set_friction(self):
        h = Mesh.open(self.hgrid)
        fric = h.set_friction('manning', 0.025)
        self.assertIsInstance(fric, Fgrid)

    def test_set_tides_boundary_forcing(self):
        h = Mesh.open(self.hgrid)
        from pyschism.forcing import Tides
        h.hgrid.generate_boundaries()
        h.set_boundary_forcing(Tides())

    def test_set_tides_boundary_forcing_by_id(self):
        h = Mesh.open(self.hgrid)
        from pyschism.forcing import Tides
        h.hgrid.generate_boundaries()
        h.set_boundary_forcing(Tides(), 0)

    def test_crs(self):
        h = Mesh.open(self.hgrid)
        h.crs

    def test_raise_ics_no_projection(self):
        h = Mesh.open(self.hgrid)
        self.assertRaises(Exception, getattr, h, "ics")

    def test_ics_latlon(self):
        h = Mesh.open(self.hgrid, crs=4326)
        self.assertEqual(h.ics, 2)

    def test_ics_meters(self):
        h = Mesh.open(self.hgrid, crs=3395)
        self.assertEqual(h.ics, 1)

    def test_slma0_sfea0(self):
        h = Mesh.open(self.hgrid, crs=3395)
        h.slam0, h.sfea0


if __name__ == '__main__':
    unittest.main()



from datetime import datetime, timedelta
from enum import Enum
import pathlib
import tempfile
from typing import Union
# import warnings

import pytz
from matplotlib.transforms import Bbox  # type: ignore[import]
from netCDF4 import Dataset  # type: ignore[import]
import numpy as np  # type: ignore[import]

from pyschism.enums import GFSProduct
from pyschism.forcing.atmosphere.nws.nws2.sflux import (
    SfluxDataset,
    AirComponent,
    PrcComponent,
    RadComponent,
)

BASE_URL = 'https://nomads.ncep.noaa.gov/dods'


class BaseURL(Enum):
    GFS_0P25 = f'{BASE_URL}/gfs_0p25'
    GFS_0P25_1HR = f'{BASE_URL}/gfs_0p25_1hr'
    GFS_0P50 = f'{BASE_URL}/gfs_0p50'
    GFS_1P00 = f'{BASE_URL}/gfs_1p00'

    @classmethod
    def _missing_(self, name):
        raise ValueError(f'{name} is not a valid GFS product.')


class Tmpdir:

    def __get__(self, obj, val):
        tmpdir = obj.__dict__.get('tmpdir')
        if tmpdir is None:
            tmpdir = tempfile.TemporaryDirectory()
            obj.__dict__['tmpdir'] = tmpdir
            obj.resource = tmpdir.name
        return tmpdir


class StartDate:

    def __set__(self, obj, start_date):
        if start_date is not None:
            obj.__dict__['start_date'] = start_date

    def __get__(self, obj, val):
        start_date = obj.__dict__.get(
            'start_date', obj.utc_timezone.localize(datetime.utcnow()))
        return datetime(start_date.year, start_date.month, start_date.day,
                        int(6 * np.floor(start_date.hour/6)),
                        tzinfo=start_date.tzinfo).astimezone(obj.utc_timezone)


class Inventory:

    def __get__(self, obj, val):
        utc_now = obj.utc_timezone.localize(datetime.utcnow())
        current_cycle = int(6 * np.floor(utc_now.hour/6))
        nomads_now = obj.utc_timezone.localize(datetime(
            utc_now.year, utc_now.month, utc_now.day, current_cycle))
        nomads_start_dates = [nomads_now - i*obj.forecast_interval
                              for i in range(40)]
        inventory = {}
        for reference_date in nomads_start_dates:
            nearest_cycle = int(6 * np.floor(reference_date.hour/6))
            if nearest_cycle != 0:
                # TODO: SCHISM is limited to cycle 00z only.
                continue
            if reference_date + timedelta(hours=0.5) > obj.start_date:
                self.add_dataset_to_inventory(obj, inventory, reference_date)
            else:
                self.add_dataset_to_inventory(obj, inventory, reference_date)
                break
        return inventory

    def add_dataset_to_inventory(self, obj, inventory, reference_date):
        _reference_date = reference_date.strftime('%Y%m%d')
        start_data_url = obj.base_url + f'/gfs{_reference_date}'
        nearest_cycle = int(6 * np.floor(reference_date.hour/6))
        fname = f'{obj.product.value}_{nearest_cycle:02d}z'
        file_url = start_data_url + f'/{fname}'
        try:
            inventory[reference_date] = Dataset(file_url)
        except OSError as e:
            if e.errno == -70:
                cdate = reference_date - obj.forecast_interval
                closest_cycle = int(6 * np.floor(cdate.hour/6))
                nomads_closest = obj.utc_timezone.localize(datetime(
                    cdate.year, cdate.month, cdate.day, closest_cycle))
                if reference_date not in inventory.keys():
                    _reference_date = nomads_closest.strftime('%Y%m%d')
                    start_data_url = obj.base_url + f'/gfs{_reference_date}'
                    fname = f'{obj.product.value}_' \
                            f'{nomads_closest.hour:02d}z'
                    file_url = start_data_url + f'/{fname}'
                    inventory[cdate] = Dataset(file_url)
            else:
                raise e


class GlobalForecastSystem(SfluxDataset):

    _tmpdir = Tmpdir()
    _inventory = Inventory()
    start_date = StartDate()

    def __init__(
        self,
        product: Union[str, GFSProduct] = GFSProduct.GFS_0P25_1HR,
    ):

        if not isinstance(product, GFSProduct):
            product = GFSProduct(product)
        self.product = product
        self.base_url = BaseURL[product.name].value
        self.prmsl_name = 'prmslmsl'
        self.spfh_name = 'spfh2m'
        self.stmp_name = 'tmpsfc'
        self.uwind_name = 'ugrd10m'
        self.vwind_name = 'vgrd10m'
        self.prate_name = 'pratesfc'
        self.dlwrf_name = 'dlwrfsfc'
        self.dswrf_name = 'dswrfsfc'

    def fetch_data(
            self,
            start_date: datetime = None,
            rnday: Union[float, timedelta] = None,
            air: bool = True,
            prc: bool = True,
            rad: bool = True,
            bbox: Bbox = None,
            timezone=None,
    ):
        """Fetches GFS data from NOMADS server. """

        self.start_date = start_date
        self._rnday = rnday
        # inventory = self.get_inventory(start_date, rnday)
        if timezone is None:
            timezone = self.utc_timezone
        if len(self._inventory) == 0:
            raise ValueError("No data. Perhaps the server is down?")
        for date in sorted(self._inventory.keys()):
            date = date.astimezone(timezone)
            nc = self._inventory[date]
            lons = nc.variables['lon'][:]
            lats = nc.variables['lat'][:]
            if bbox is None:
                bbox = Bbox.from_extents(0, -90, 360, 90)
            else:
                xmin = bbox.xmin + 360 if bbox.xmin < 0 else bbox.xmin
                xmax = bbox.xmax + 360 if bbox.xmax < 0 else bbox.xmax
                bbox = Bbox.from_extents(xmin, bbox.ymin, xmax, bbox.ymax)

            lat_idxs = np.where((lats >= bbox.ymin)
                                & (lats <= bbox.ymax))[0]
            lats = lats[lat_idxs]
            lon_idxs = np.where((lons >= bbox.xmin)
                                & (lons <= bbox.xmax))[0]
            _lons = []
            for x in lons[lon_idxs]:
                if x > 180:
                    _lons.append(x-360)
                else:
                    _lons.append(x)
            lons = np.array(_lons)
            x, y = np.meshgrid(lons, lats)
            if air is True:
                with Dataset(pathlib.Path(self._tmpdir.name) /
                             f"air_{self.product.value}_{date}.nc", 'w',
                             format='NETCDF3_CLASSIC') as dst:
                    # global attributes
                    dst.setncatts({"Conventions": "CF-1.0"})
                    # dimensions
                    dst.createDimension('nx_grid', lons.shape[0])
                    dst.createDimension('ny_grid', lats.shape[0])
                    dst.createDimension('time', None)
                    # variables
                    # lon
                    dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
                    dst['lon'].long_name = "Longitude"
                    dst['lon'].standard_name = "longitude"
                    dst['lon'].units = "degrees_east"
                    dst['lon'][:] = x
                    # lat
                    dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
                    dst['lat'].long_name = "Latitude"
                    dst['lat'].standard_name = "latitude"
                    dst['lat'].units = "degrees_north"
                    dst['lat'][:] = y
                    # time
                    dst.createVariable('time', 'f4', ('time',))
                    dst['time'].long_name = 'Time'
                    dst['time'].standard_name = 'time'
                    dst['time'].units = f'days since {date.year}-{date.month}'\
                                        f'-{date.day}'
                    dst['time'].base_date = (date.year, date.month, date.day,
                                             date.hour)
                    dst['time'][:] = np.arange(
                        nc['time'].resolution,
                        nc['time'].resolution*(nc['time'].shape[0]+1),
                        step=nc['time'].resolution)

                    # prmsl
                    dst.createVariable(
                        'prmsl', nc.variables[self.prmsl_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['prmsl'].long_name = "Pressure reduced to MSL"
                    dst['prmsl'].standard_name = "air_pressure_at_sea_level"
                    dst['prmsl'].units = "Pa"
                    dst['prmsl'][:] = nc.variables[
                        self.prmsl_name][:, lat_idxs, lon_idxs]

                    # spfh
                    dst.createVariable(
                        'spfh', nc.variables[self.spfh_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['spfh'].long_name = "Surface Specific Humidity "\
                                            "(2m AGL)"
                    dst['spfh'].standard_name = "specific_humidity"
                    dst['spfh'].units = "1"
                    dst['spfh'][:] = nc.variables[
                        self.spfh_name][:, lat_idxs, lon_idxs]

                    # stmp
                    dst.createVariable(
                        'stmp', nc.variables[self.stmp_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['stmp'].long_name = "Surface Air Temperature (2m AGL)"
                    dst['stmp'].standard_name = "air_temperature"
                    dst['stmp'].units = "K"
                    dst['stmp'][:] = nc.variables[
                        self.stmp_name][:, lat_idxs, lon_idxs]

                    # uwind
                    dst.createVariable(
                        'uwind', nc.variables[self.uwind_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['uwind'].long_name = "Surface Eastward Air Velocity "\
                                             "(10m AGL)"
                    dst['uwind'].standard_name = "eastward_wind"
                    dst['uwind'].units = "m/s"
                    dst['uwind'][:] = nc.variables[
                        self.uwind_name][:, lat_idxs, lon_idxs]

                    # vwind
                    dst.createVariable(
                        'vwind', nc.variables[self.vwind_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['vwind'].long_name = "Surface Northward Air Velocity "\
                                             "(10m AGL)"
                    dst['vwind'].standard_name = "northward_wind"
                    dst['vwind'].units = "m/s"
                    dst['vwind'][:] = nc.variables[
                        self.vwind_name][:, lat_idxs, lon_idxs]

            if prc is True:
                with Dataset(pathlib.Path(self._tmpdir.name) /
                             f"prc_{self.product.value}_{date}.nc", 'w',
                             format='NETCDF3_CLASSIC') as dst:
                    # global attributes
                    dst.setncatts({"Conventions": "CF-1.0"})
                    # dimensions
                    dst.createDimension('nx_grid', lons.shape[0])
                    dst.createDimension('ny_grid', lats.shape[0])
                    dst.createDimension('time', None)
                    # variables
                    # lon
                    dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
                    dst['lon'].long_name = "Longitude"
                    dst['lon'].standard_name = "longitude"
                    dst['lon'].units = "degrees_east"
                    dst['lon'][:] = x
                    # lat
                    dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
                    dst['lat'].long_name = "Latitude"
                    dst['lat'].standard_name = "latitude"
                    dst['lat'].units = "degrees_north"
                    dst['lat'][:] = y
                    # time
                    dst.createVariable('time', 'f4', ('time',))
                    dst['time'].long_name = 'Time'
                    dst['time'].standard_name = 'time'
                    dst['time'].units = f'days since {date.year}-{date.month}'\
                                        f'-{date.day} {date.hour:02d}:'\
                                        f"{date.minute:02d} " \
                                        f'{date.tzinfo}'
                    dst['time'].base_date = (date.year, date.month, date.day,
                                             date.hour)
                    dst['time'][:] = np.arange(
                        nc['time'].resolution,
                        nc['time'].resolution*(len(nc['time'][:])),
                        step=nc['time'].resolution)

                    # prate
                    dst.createVariable(
                        'prate', nc.variables[self.prate_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['prate'].long_name = "Surface Precipitation Rate"
                    dst['prate'].standard_name = "air_pressure_at_sea_level"
                    dst['prate'].units = "kg/m^2/s"
                    dst['prate'][:] = nc.variables[
                        self.prate_name][:, lat_idxs, lon_idxs]

            if rad is True:
                with Dataset(pathlib.Path(self._tmpdir.name) /
                             f"rad_{self.product.value}_{date}.nc", 'w',
                             format='NETCDF3_CLASSIC') as dst:
                    # global attributes
                    dst.setncatts({"Conventions": "CF-1.0"})
                    # dimensions
                    dst.createDimension('nx_grid', lons.shape[0])
                    dst.createDimension('ny_grid', lats.shape[0])
                    dst.createDimension('time', None)
                    # variables
                    # lon
                    dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
                    dst['lon'].long_name = "Longitude"
                    dst['lon'].standard_name = "longitude"
                    dst['lon'].units = "degrees_east"
                    dst['lon'][:] = x
                    # lat
                    dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
                    dst['lat'].long_name = "Latitude"
                    dst['lat'].standard_name = "latitude"
                    dst['lat'].units = "degrees_north"
                    dst['lat'][:] = y
                    # time
                    dst.createVariable('time', 'f4', ('time',))
                    dst['time'].long_name = 'Time'
                    dst['time'].standard_name = 'time'
                    dst['time'].units = f'days since {date.year}-{date.month}'\
                                        f'-{date.day}T{date.hour:02d}:'\
                                        f'{date.minute:02d}'\
                                        f'{date.tzinfo}'
                    dst['time'].base_date = (date.year, date.month, date.day,
                                             date.hour)
                    dst['time'][:] = np.arange(
                        nc['time'].resolution,
                        nc['time'].resolution*(len(nc['time'][:])),
                        step=nc['time'].resolution)

                    # dlwrf
                    dst.createVariable(
                        'dlwrf', nc.variables[self.dlwrf_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['dlwrf'].long_name = "Downward Long Wave Radiation "\
                                             "Flux"
                    dst['dlwrf'].standard_name = "surface_downwelling_"\
                                                 "longwave_flux_in_air"
                    dst['dlwrf'].units = "W/m^2"
                    dst['dlwrf'][:] = nc.variables[
                        self.dlwrf_name][:, lat_idxs, lon_idxs]

                    # dswrf
                    dst.createVariable(
                        'dswrf', nc.variables[self.dswrf_name].datatype,
                        ('time', 'ny_grid', 'nx_grid'))
                    dst['dswrf'].long_name = "Downward Short Wave Radiation "\
                                             "Flux"
                    dst['dswrf'].standard_name = "surface_downwelling_"\
                                                 "shortwave_flux_in_air"
                    dst['dswrf'].units = "W/m^2"
                    dst['dswrf'][:] = nc.variables[
                        self.dswrf_name][:, lat_idxs, lon_idxs]

        self.resource = self._tmpdir.name
        if air:
            self.air = AirComponent(self.fields)
        if prc:
            self.prc = PrcComponent(self.fields)
        if rad:
            self.rad = RadComponent(self.fields)

    @property
    def forecast_interval(self):
        return timedelta(hours=6)

    @property
    def utc_timezone(self):
        return pytz.timezone('UTC')

    @property
    def nomads_server_ttl(self):
        return timedelta(days=10)



from collections import defaultdict
import os
import numbers
import pathlib
from typing import Union, Dict, TextIO
import warnings

import numpy as np  # type: ignore[import]
from pyproj import CRS  # type: ignore[import]
from pyproj.exceptions import CRSError  # type: ignore[import]


def buffer_to_dict(buf: TextIO):
    description = buf.readline()
    NE, NP = map(int, buf.readline().split())
    nodes = {}
    for _ in range(NP):
        line = buf.readline().strip('\n').split()
        # Gr3/fort.14 format cannot distinguish between a 2D mesh with one
        # vector value (e.g. velocity, which uses 2 columns) or a 3D mesh with
        # one scalar value. This is a design problem of the mesh format, which
        # renders it ambiguous, and the main reason why the use of fort.14/grd
        # formats is discouraged, in favor of UGRID.
        # Here, we assume the input mesh is strictly a 2D mesh, and the data
        # that follows is an array of values.
        if len(line[3:]) == 1:
            nodes[line[0]] = [
                (float(line[1]), float(line[2])), float(line[3])]
        else:
            nodes[line[0]] = [
                (float(line[1]), float(line[2])),
                [float(line[i]) for i in range(3, len(line[3:]))]
            ]
    elements = {}
    for _ in range(NE):
        line = buf.readline().split()
        elements[line[0]] = line[2:]
    # Assume EOF if NOPE is empty.
    try:
        NOPE = int(buf.readline().split()[0])
    except IndexError:
        return {'description': description,
                'nodes': nodes,
                'elements': elements}
    # let NOPE=-1 mean an ellipsoidal-mesh
    # reassigning NOPE to 0 until further implementation is applied.
    boundaries: Dict = defaultdict(dict)
    _bnd_id = 0
    buf.readline()
    while _bnd_id < NOPE:
        NETA = int(buf.readline().split()[0])
        _cnt = 0
        boundaries[None][_bnd_id] = dict()
        boundaries[None][_bnd_id]['indexes'] = list()
        while _cnt < NETA:
            boundaries[None][_bnd_id]['indexes'].append(
                buf.readline().split()[0].strip())
            _cnt += 1
        _bnd_id += 1
    NBOU = int(buf.readline().split()[0])
    _nbnd_cnt = 0
    buf.readline()
    while _nbnd_cnt < NBOU:
        npts, ibtype = map(int, buf.readline().split()[:2])
        _pnt_cnt = 0
        if ibtype not in boundaries:
            _bnd_id = 0
        else:
            _bnd_id = len(boundaries[ibtype])
        boundaries[ibtype][_bnd_id] = dict()
        boundaries[ibtype][_bnd_id]['indexes'] = list()
        while _pnt_cnt < npts:
            line = buf.readline().split()
            if len(line) == 1:
                boundaries[ibtype][_bnd_id]['indexes'].append(line[0])
            else:
                index_construct = []
                for val in line:
                    if '.' in val:
                        continue
                    index_construct.append(val)
                boundaries[ibtype][_bnd_id]['indexes'].append(index_construct)
            _pnt_cnt += 1
        _nbnd_cnt += 1
    return {'description': description,
            'nodes': nodes,
            'elements': elements,
            'boundaries': boundaries}


def to_string(description, nodes, elements, boundaries=None, crs=None):
    """
    must contain keys:
        description
        vertices
        elements
        vertex_id
        element_id
        values
        boundaries (optional)
            indexes
    """
    NE, NP = len(elements), len(nodes)
    out = [f"{description}", f"{NE} {NP}"]
    # TODO: Probably faster if using np.array2string
    for id, (coords, values) in nodes.items():
        if isinstance(values, numbers.Number):
            values = [values]
        line = [f"{id}"]
        line.extend([f"{x:<.16E}" for x in coords])
        line.extend([f"{x:<.16E}" for x in values])
        out.append(" ".join(line))

    for id, element in elements.items():
        line = [f"{id}"]
        line.append(f"{len(element)}")
        line.extend([f"{e}" for e in element])
        out.append(" ".join(line))

    # ocean boundaries
    if boundaries is not None:
        out.append(f"{len(boundaries[None]):d} "
                   "! total number of ocean boundaries")
        # count total number of ocean boundaries
        _sum = 0
        for bnd in boundaries[None].values():
            _sum += len(bnd['indexes'])
        out.append(f"{int(_sum):d} ! total number of ocean boundary nodes")
        # write ocean boundary indexes
        for i, boundary in boundaries[None].items():
            out.append(f"{len(boundary['indexes']):d}"
                       f" ! number of nodes for ocean_boundary_{i}")
            for idx in boundary['indexes']:
                out.append(f"{idx}")
    else:
        out.append("0 ! total number of ocean boundaries")
        out.append("0 ! total number of ocean boundary nodes")
    # remaining boundaries
    boundaries = {} if boundaries is None else boundaries
    _cnt = 0
    for key in boundaries:
        if key is not None:
            for bnd in boundaries[key]:
                _cnt += 1
    out.append(f"{_cnt:d}  ! total number of non-ocean boundaries")
    # count remaining boundary nodes
    _cnt = 0
    for ibtype in boundaries:
        if ibtype is not None:
            for bnd in boundaries[ibtype].values():
                _cnt += np.asarray(bnd['indexes']).size
    out.append(f"{_cnt:d} ! Total number of non-ocean boundary nodes")
    # all additional boundaries
    for ibtype, boundaries in boundaries.items():
        if ibtype is None:
            continue
        for id, boundary in boundaries.items():
            line = [
                f"{len(boundary['indexes']):d}",
                f"{ibtype}",
                f"! boundary {ibtype}:{id}"]
            out.append(' '.join(line))
            for idx in boundary['indexes']:
                out.append(f"{idx}")
    return "\n".join(out)


def read(resource: Union[str, os.PathLike], boundaries: bool = True, crs=True):
    """Converts a file-like object representing a grd-formatted unstructured
    mesh into a python dictionary:

    Args:
        resource: Path to file on disk or file-like object such as
            :class:`io.StringIO`
    """
    resource = pathlib.Path(resource)
    with open(resource, 'r') as stream:
        grd = buffer_to_dict(stream)
    if boundaries is False:
        grd.pop('boundaries', None)
    if crs is True:
        crs = None
    if crs is None:
        for try_crs in grd['description'].split():
            try:
                crs = CRS.from_user_input(try_crs)
                break
            except CRSError:
                pass
    if crs is None:
        warnings.warn(f'File {str(resource)} does not contain CRS '
                      'information and no CRS was given.')
    if crs is not False:
        grd.update({'crs': crs})
    return grd


def write(grd, path, overwrite=False):
    path = pathlib.Path(path)
    if path.is_file() and not overwrite:
        raise Exception('File exists, pass overwrite=True to allow overwrite.')
    with open(path, 'w') as f:
        f.write(to_string(**grd))



import pathlib

from matplotlib.animation import FuncAnimation
import matplotlib.pyplot as plt
from matplotlib.tri import Triangulation
from netCDF4 import Dataset
import numpy as np
from pyugrid import UGrid

from pyschism.enums import OutputVariableUnit, OutputVariableShortName


class PlotOutputCombined:

    def __init__(self, path):
        self.nc = Dataset(pathlib.Path(path))

    def plot(self, variable, show=False, index=None):
        if index is None:
            self.animation(
                variable,
                # show=True,
                save='/home/jreniel/pyschism/examples/example_1/test.gif',
                vmin=0,
                vmax=3,
                # start_frame=200,
                # end_frame=300,
                )
        else:
            var = self.nc[variable]
            ugrid = UGrid.from_nc_dataset(self.nc)
            x = ugrid.nodes[:, 0]
            y = ugrid.nodes[:, 1]
            triangulation = Triangulation(x, y, ugrid.faces[:, :3])
            triangulation.set_mask(self.nc['wetdry_elem'][index])
            plt.tricontourf(
                triangulation, var[index, :], levels=256, cmap='jet')
            plt.gca().axis('scaled')
            if show:
                plt.show()

    def animation(
            self,
            variable,
            save=False,
            fps=3,
            start_frame=0,
            end_frame=-1,
            figsize=None,
            wireframe=False,
            cmap='jet',
            levels=256,
            show=False,
            xmin=None,
            xmax=None,
            ymin=None,
            ymax=None,
            vmin=None,
            vmax=None,

    ):

        fig = plt.figure(figsize)
        ax = fig.add_subplot(111)

        plt.tight_layout(pad=2)

        ugrid = UGrid.from_nc_dataset(self.nc)
        x = ugrid.nodes[:, 0]
        y = ugrid.nodes[:, 1]
        triangulation = Triangulation(x, y, ugrid.faces[:, :3])
        xmin = np.min(x) if xmin is None else xmin
        xmax = np.max(x) if xmax is None else xmax
        ymin = np.min(y) if ymin is None else ymin
        ymax = np.max(y) if ymax is None else ymax
        vmin = np.min(self.nc[variable]) if vmin is None else vmin
        vmax = np.max(self.nc[variable]) if vmax is None else vmax
        unit = OutputVariableUnit[OutputVariableShortName(variable).name].value

        def animate(index):
            _ax = fig.get_axes()
            ax.clear()
            if len(_ax) > 1:
                cax = _ax[1]
                cax.cla()
            else:
                cax = None

            triangulation.set_mask(self.nc['wetdry_elem'][index])

            if wireframe:
                ax.triplot(triangulation, color='k', linewidth=0.7)

            ax.tricontourf(
                triangulation,
                self.nc[variable][index, :],
                cmap=cmap,
                levels=levels,
                vmin=vmin,
                vmax=vmax
                )

            ax.set_ylim(ymin, ymax, auto=True)
            ax.set_xlim(xmin, xmax, auto=True)

            ax.set_xlabel('Longitude (°E)')
            ax.set_ylabel('Latitude (°N)')

            # ax.set_title(dates[i].strftime('%b %d, %Y %H:%M'))
            m = plt.cm.ScalarMappable(cmap=cmap)
            m.set_array(self.nc[variable][index, :])
            m.set_clim(vmin, vmax)
            cbar = fig.colorbar(m, cax=cax, format='%.1f', boundaries=np.linspace(vmin, vmax, levels))

            # cbar = fig.colorbar(_ax)
            cbar.ax.set_ylabel(f'{variable} [{unit}]', rotation=90)

        end_frame = end_frame % self.nc[variable].shape[0] \
            if end_frame < 0 else end_frame
        start_frame = start_frame % self.nc[variable].shape[0] \
            if start_frame < 0 else start_frame
        frames = range(start_frame, end_frame)
        anim = FuncAnimation(
            fig,
            animate,
            frames,
            blit=False
            )

        if save:
            anim.save(
                pathlib.Path(save),
                writer='imagemagick',
                fps=fps
            )

        if show:
            plt.show()

        return anim



from psutil import cpu_count  # type: ignore[import]


class ServerConfig:

    def __init__(
            self,
            nproc: int = None,
            symlink_outputs: str = None,
            schism_binary: str = None,
            mpi_launcher: str = None,
    ):
        self.nproc = nproc
        self.symlink_outputs = symlink_outputs
        self.schism_binary = schism_binary
        self.mpi_launcher = mpi_launcher

    def __str__(self):
        f = []
        f.append(self.MPI_LAUNCHER)
        f.append(self.SCHISM_BINARY)
        if self.symlink_outputs is not None:
            f.append(self.SYMLINK_OUTPUTS)
        return "\n".join(f)

    @property
    def mpi_launcher(self):
        return self.__mpi_launcher

    @mpi_launcher.setter
    def mpi_launcher(self, mpi_launcher):
        self.__mpi_launcher = "mpiexec -n" if mpi_launcher is None \
            else mpi_launcher

    @property
    def nproc(self):
        return self.__nproc

    @nproc.setter
    def nproc(self, nproc):
        self.__nproc = cpu_count(logical=False) if nproc is None else nproc

    @property
    def SCHISM_BINARY(self):
        f = "SCHISM_BINARY="
        if self.schism_binary is None:
            f += "pschism_TVD-VL"
        return f

    @property
    def SYMLINK_OUTPUTS(self):
        f = "SYMLINK_OUTPUTS="
        if self.symlink_outputs is not None:
            f += f"{self.symlink_outputs}"
        return f

    @property
    def MPI_LAUNCHER(self):
        f = "MPI_LAUNCHER="
        if self.mpi_launcher is not None:
            f += f'mpiexec -n {self.nproc}'
        return f



from datetime import datetime, timedelta
import logging
import pathlib
from typing import Union

import f90nml
import pytz

from pyschism.enums import Coriolis
from pyschism.forcing.atmosphere.nws import NWS
from pyschism.forcing.atmosphere.nws.nws2 import NWS2

PARAM_TEMPLATE = pathlib.Path(__file__).parent / 'param.nml.template'
PARAM_DEFAULTS = f90nml.read(PARAM_TEMPLATE)['opt']

_logger = logging.getLogger(__name__)

class OptMeta(type):

    def __new__(meta, name, bases, attrs):
        for key, value in PARAM_DEFAULTS.items():
            if key not in attrs:
                if isinstance(value, list):
                    attrs[key] = len(value)*[0]
                else:
                    attrs[key] = None
        return type(name, bases, attrs)


class Dramp:

    def __set__(self, obj, dramp: Union[int, float, timedelta, None]):

        if not isinstance(dramp, (int, float, timedelta, type(None))):
            raise TypeError("Argument drampbc must be an int, float, "
                            "timedelta, or None.")

        if isinstance(dramp, (int, float)):
            dramp = timedelta(days=dramp)

        if dramp is not None:
            obj.nramp = 1

        obj.__dict__['dramp'] = dramp

    def __get__(self, obj, val):
        return obj.__dict__.get('dramp')


class Drampbc:

    def __set__(self, obj, drampbc: Union[int, float, timedelta, None]):
        if not isinstance(drampbc, (int, float, timedelta, type(None))):
            raise TypeError("Argument drampbc must be an int, float, "
                            "timedelta, or None.")

        if isinstance(drampbc, (int, float)):
            drampbc = timedelta(days=drampbc)

        if drampbc is not None:
            obj.nrampbc = 1

        obj.__dict__['drampbc'] = drampbc

    def __get__(self, obj, val):
        return obj.__dict__.get('drampbc')


class StartDate:

    def __set__(self, obj, start_date: Union[datetime, None]):
        if not isinstance(start_date, (datetime, type(None))):
            raise TypeError("Argument start_date must be of type datetime or "
                            "None. The datetime object will be assumed to be "
                            "in UTC if ")
        if start_date is not None:
            if start_date.tzinfo is None \
                    or start_date.tzinfo.utcoffset(start_date) is None:
                start_date = start_date.replace(tzinfo=pytz.utc)
            obj.start_year = start_date.year
            obj.start_month = start_date.month
            obj.start_day = start_date.day
            obj.start_hour = start_date.hour
            obj.start_hour += start_date.minute / 60.
            obj.utc_start = -start_date.utcoffset().total_seconds() / 3600  # type: ignore[union-attr]  # noqa: E501
            # get rid of "negative" zero
            obj.utc_start = +0. if obj.utc_start == -0. \
                else obj.utc_start
        obj.__dict__['start_date'] = start_date

    def __get__(self, obj, val):
        return obj.__dict__.get('start_date')


class Nchi:

    def __set__(self, obj, fgrid):
        nchi = fgrid.nchi
        if nchi == -1:
            obj.hmin_man = fgrid.hmin_man
        if obj.nchi == 1:
            obj.dbz_min = fgrid.dbz_min
            obj.dbz_decay = fgrid.dbz_decay
        obj.__dict__['nchi'] = nchi

    def __get__(self, obj, val):
        return obj.__dict__.get('nchi')


class Ics:

    def __set__(self, obj, ics: int):
        obj.__dict__['ics'] = ics

    def __get__(self, obj, val):
        return obj.__dict__.get('ics')


class Sfea0:

    def __set__(self, obj, sfea0: float):
        obj.__dict__['sfea0'] = sfea0

    def __get__(self, obj, val):
        return obj.__dict__.get('sfea0')


class Ncor:

    def __set__(self, obj, ncor: Coriolis):
        if not isinstance(ncor, Coriolis):
            raise TypeError(f"ncor must be of type {Coriolis}, not type "
                            f"{type(ncor)}.")
        obj.__dict__['ncor'] = ncor.value

    def __get__(self, obj, val):
        return obj.__dict__.get('ncor')


class Ihot:

    def __set__(self, obj, ihot: int):
        if not isinstance(ihot, int):
            raise TypeError(f"ihot must be of type {int}, not type "
                            f"{type(ihot)}.")
        if ihot not in [0, 1]:
            raise ValueError("ihot must be 0 or 1")

        obj.__dict__['ihot'] = ihot

    def __get__(self, obj, val):
        return obj.__dict__.get('ihot')


class Nws:
    def __set__(self, obj, nws: NWS):
        if not isinstance(nws, NWS):
            raise TypeError(
                f"nws must be of type {NWS}, not type {type(nws)}.")

        if obj.start_date is None:
            raise ValueError(
                "Can't initialize atmospheric data without start_date")
            nws._start_date = obj.start_date

        if isinstance(nws, NWS2):
            obj.__dict__['nws'] = 2

        else:
            raise NotImplementedError(
                f'NWS type {type(nws)} is not implemented.')

    def __get__(self, obj, val):
        return obj.__dict__.get('nws')


class OPT(metaclass=OptMeta):
    """ Provides error checking implementation for OPT group """

    dramp = Dramp()
    drampbc = Drampbc()
    start_date = StartDate()
    nchi = Nchi()
    ics = Ics()
    ncor = Ncor()
    sfea0 = Sfea0()
    ihot = Ihot()
    nws = Nws()

    def __init__(
            self,
            dramp: Union[int, float, timedelta] = None,
            drampbc: Union[int, float, timedelta] = None,
            # drampwind:,
            start_date: datetime = None):
        _logger.info('Iniatializing OPT.')
        self.dramp = dramp
        self.drampbc = drampbc
        self.start_date = start_date
        # TODO: Generate descriptors for the following properties:
        # wtiminc = 150. !time step for atmos. forcing. Default: same as dt
        # nrampwind = 1 !ramp-up option for atmos. forcing
        # drampwind = 1. !needed if nrampwind/=0; ramp-up period in days
        # iwindoff = 0 !needed only if nws/=0; '1': needs windfactor.gr3
        # iwind_form = -1 !needed if nws/=0
        # impose_net_flux = 0

    def __str__(self):
        data = []
        for key, default in PARAM_DEFAULTS.items():
            current = getattr(self, key)
            if key in ['dramp', 'drampbc']:
                if current is not None:
                    current = current.total_seconds() / (60*60*24)
            if current is not None:
                if not isinstance(current, list):
                    data.append(f'  {key}={current}')
                else:
                    for i, state in enumerate(current):
                        if state:
                            current.append(f'  {current}({i+1}) = 1')
        data = '\n'.join(data)
        return f"&OPT\n{data}\n/"



import logging
import os
import pathlib
from typing import Union

import pandas as pd


_logger = logging.getLogger(__name__)


class TimeHistoryFile:

    def __init__(self, df, start_date, rnday, filename):
        self.df = df
        self.filename = filename
        self.start_date = start_date
        self.rnday = rnday

    def write(self, path: Union[str, os.PathLike], overwrite: bool = False):
        elements = self.df['element_id'].unique()
        times = self.df['time'].unique()
        with open(pathlib.Path(path) / self.filename, 'w') as f:
            for time in times:
                relative_time = (time - self.start_date).total_seconds()
                if relative_time < 0:
                    continue
                line = [f"{relative_time:G}"]
                for element_id in elements:
                    pd_series = self.df["flow"][
                        (self.df["time"] == time) &
                        (self.df["element_id"] == element_id)]
                    line.append(f'{pd_series.values[0]:.4e}')
                f.write(' '.join(line))
                f.write('\n')


class Vsource(TimeHistoryFile):

    def __init__(self, df, start_date, rnday, filename='vsource.th'):
        super().__init__(df, start_date, rnday, filename)


class Vsink(TimeHistoryFile):

    def __init__(self, df, start_date, rnday, filename='vsink.th'):
        super().__init__(df, start_date, rnday, filename)


class Msource(TimeHistoryFile):

    def __init__(self, df, start_date, rnday, filename='msource.th'):
        super().__init__(df, start_date, rnday, filename)

    def write(self, path: Union[str, os.PathLike], overwrite: bool = False):
        elements = self.df['element_id'].unique()
        times = self.df['time'].unique()
        with open(pathlib.Path(path) / self.filename, 'w') as f:
            for time in times:
                relative_time = (time - self.start_date).total_seconds()
                if relative_time < 0:
                    continue
                line = [f"{relative_time:G}"]
                for element_id in elements:
                    pd_series = self.df["temperature"][
                        (self.df["time"] == time) &
                        (self.df["element_id"] == element_id)]
                    line.append(f'{pd_series.values[0]:.4e}')
                for element_id in elements:
                    pd_series = self.df["salinity"][
                        (self.df["time"] == time) &
                        (self.df["element_id"] == element_id)]
                    line.append(f'{pd_series.values[0]:.4e}')
                f.write(' '.join(line))
                f.write('\n')


class SourcesDataFrame:
    '''This descriptor holds the shared dataframe for the API components.'''

    def __get__(self, obj, val):
        if not hasattr(self, '_df'):
            self._df = pd.DataFrame(
                columns=[
                    'element_id',
                    'time',
                    'flow',
                    'temperature',
                    'salinity'
                ])
        return self._df

    def __set__(self, obj, df: pd.DataFrame):
        self._df = df


class Sources:
    '''Public interface used for adding sources to the model.'''

    _df = SourcesDataFrame()

    def __len__(self):
        return len(self.df)

    def add_data(self, time, element_id, flow, temperature, salinity):

        # TODO: sources/sinks.add_data is slow, it should be vectorized
        # main reason is because it's tallying multiple datapoints that
        # corresponds to the same element, so it has to check the dataframe
        # for each datapoint added.

        # TODO: Consider additional inputs besides TS?
        # Check if datapoint already exists (auto-tally)
        tally = ((self.df['time'] == time) &
                 (self.df['element_id'] == element_id))
        if tally.any():
            # previous_flow = self.df.loc[tally, 'flow']
            self.df.loc[tally, 'flow'] += flow
            # TODO: What happens if we have two different flows that both are
            # assigned to the same element? Example: 100 m^3/s @ 1 psu then
            # another flow on the same element of 1 m^3/s @ 100 psu. How do we
            # combine these on a single element? Flow is just simple summation,
            # but TS is different.
            if (self.df.loc[tally, 'salinity'] != salinity).any():
                raise NotImplementedError('Two different values of salinity '
                                          'for same time/element')

            if (self.df.loc[tally, 'temperature'] != temperature).any():
                raise NotImplementedError('Two different values of temperature'
                                          ' for same time/element')

        else:
            self._df = self.df.append(
                pd.Series(
                    [element_id, time, flow, temperature, salinity],
                    index=self.df.columns,
                ),
                ignore_index=True,
                # TODO: in_place=True, ?
            )

    @property
    def df(self):
        return self._df


class SinksDataFrame:
    '''This descriptor holds the shared dataframe for the API components.'''

    def __get__(self, obj, val):
        if not hasattr(self, '_df'):
            self._df = pd.DataFrame(
                columns=[
                    'element_id',
                    'time',
                    'flow'
                ])
        return self._df

    def __set__(self, obj, df: pd.DataFrame):
        self._df = df


class Sinks:

    _df = SinksDataFrame()

    def __len__(self):
        return len(self.df)

    def add_data(self, time, element_id, flow):
        tally = ((self.df['time'] == time) &
                 (self.df['element_id'] == element_id))
        if tally.any():
            self.df.loc[tally, 'flow'] += flow
        else:
            self._df = self.df.append(
                pd.Series(
                    [element_id, time, flow],
                    index=self.df.columns,
                ),
                ignore_index=True,
            )

    @property
    def df(self):
        return self._df


class SourceSink:

    def __init__(
            self,
            sources: Sources,
            sinks: Sinks,
            filename='source_sink.in'
    ):
        self._sources = sources
        self._sinks = sinks
        self.filename = filename

    def write(self, path: Union[str, os.PathLike], overwrite: bool = False):
        sources = self.sources.df['element_id'].unique()
        sinks = self.sinks.df['element_id'].unique()
        data = []
        with open(pathlib.Path(path) / self.filename, 'w') as f:
            data.append(f'{len(sources)}')
            for element_id in sources:
                data.append(f'{element_id}')
            data.append('\n')
            data.append(f'{len(sinks)}')
            for element_id in sinks:
                data.append(f'{element_id}')
            f.write('\n'.join(data))

    @property
    def sources(self):
        return self._sources

    @property
    def sinks(self):
        return self._sinks


class Hydrology:

    def __init__(self):
        self._sources = Sources()
        self._sinks = Sinks()

    def __call__(self, model_driver):
        self._model_driver = model_driver

    def write(
            self,
            path: Union[str, os.PathLike],
            overwrite: bool = False,
            msource: Union[str, bool] = True,
            vsource: Union[str, bool] = True,
            vsink: Union[str, bool] = True,
            source_sink: Union[str, bool] = True,
    ):

        start_date = self.model_driver.param.opt.start_date
        rnday = self.model_driver.param.core.rnday

        # write source sink
        if source_sink is True:
            fname = 'source_sink.in'
        elif isinstance(source_sink, str):
            fname = source_sink
        if source_sink is not False:
            SourceSink(self.sources, self.sinks, fname).write(path, overwrite)

        if vsource is True:
            fname = 'vsource.th'
        elif isinstance(vsource, str):
            fname = vsource
        if vsource is not False:
            Vsource(
                self.sources.df, start_date, rnday, fname).write(
                path, overwrite)

        if msource is True:
            fname = 'msource.th'
        elif isinstance(msource, str):
            fname = msource
        if msource is not False:
            Msource(
                self.sources.df, start_date, rnday, fname).write(
                path, overwrite)

        if vsink is True:
            fname = 'vsink.th'
        elif isinstance(vsink, str):
            fname = vsink
        if vsink is not False:
            Vsink(
                self.sinks.df, start_date, rnday, fname).write(
                path, overwrite)

    @property
    def sources(self):
        return self._sources

    @property
    def sinks(self):
        return self._sinks

    @property
    def model_driver(self):
        return self._model_driver



#! /usr/bin/env python
import pathlib
from datetime import datetime, timedelta, timezone
import logging

from pyschism.mesh import Hgrid, Vgrid, Fgrid
from pyschism import ModelDomain, ModelDriver, Stations
from pyschism.forcing import Tides



SIMPLE_SLURM_DRIVER = """#!/bin/bash --login
#SBATCH -D .
#SBATCH -J schism-test
#SBATCH -A nosofs
#SBATCH --mail-type=all
#SBATCH --mail-user=jaime.calzada@noaa.gov
#SBATCH --output=slurm.log
#SBATCH -n 500
#SBATCH --time=02:00:00
#SBATCH --partition=orion

set -e

module load intel/2020 impi/2020 netcdf/4.7.2-parallel

PATH=$HOME/SCHISM/schism/build/bin:$PATH

main() {
  mkdir -p outputs
  time srun pschism_TVD-VL
}

main
"""

# https://eev.ee/blog/2012/05/23/python-faq-descriptors/

PARENT = pathlib.Path(__file__).parent

logging.basicConfig(level=logging.INFO)

_logger = logging.getLogger(__name__)

if __name__ == '__main__':
    from time import time
    # open gr3 file
    _logger.info('Reading hgrid file...')
    _tic = time()
    hgrid = Hgrid.open(PARENT / 'hgrid.gr3',crs='EPSG:4326')
    _logger.info(f'Reading hgrind file took {time()-_tic}.')

    vgrid = Vgrid()
    fgrid = Fgrid.open(PARENT / 'manning.gr3',crs='EPSG:4326')

    # setup model domain
    domain = ModelDomain(hgrid, vgrid, fgrid)

    # set tidal boundary conditions
    elevbc = Tides()
    elevbc.use_all()  # activate all forcing constituents

    # connect the boundary condition to the domain
    domain.add_boundary_condition(elevbc)

    # from pyschism.forcing.atmosphere.nws.nws2 import NWS2
    # from pyschism.forcing.atmosphere.gfs import GlobalForecastSystem as GFS

    # sflux_1 = GFS()
    # sflux_2 = HWRF()

    # atmos = NWS2(
    #         sflux_1,
    #         sflux_2
    #     )

    # domain.set_atmospheric_forcing(atmos)

    #  ------ Param options
    # dt and rnday are required arguments

    # Use int or float for seconds, or timedelta objects for pythonic
    # specifications
    dt = timedelta(seconds=150.)
    rnday = timedelta(days=30.)

    # Use an integer for number of steps or a timedelta to approximate
    # number of steps internally based on timestep
    nspool = timedelta(minutes=30.)

    # The dramp and start_date are optional parameters.
    # start_date is required when using forcings that
    # require some natural dates, (e.g. tides, winds, T&S ...)
    # If using a forcing that requires a start date is invoked, and no start
    # date is provided, the driver will throw an exception.
    # tzinfo is optional, UTC assumed if not provided
    dramp = 0.1 * rnday
    start_date = datetime(2017, 9, 18, 4, 0,
                          tzinfo=timezone(timedelta(hours=-4))
                          ) - dramp

    # Now we add station outputs
#    stations = Stations.from_file(PARENT / 'station.in', timedelta(minutes=6.),
#                                  elev=True, u=True, v=True)

    # init the model driver
    driver = ModelDriver(domain, dt, rnday, dramp=dramp, start_date=start_date,
                         #stations=stations, nspool=nspool, elev=True)
                         nspool=nspool, elev=True)

    # Output requests, as well as most other namelist variables, can be
    # modified post instantiation through their corresponding properties.
    # For example, the following line activates the depth averaged horizontal
    # velocity output request:
    driver.param.schout.dahv = True

    # write files to disk
    outdir = pathlib.Path('staging')
    driver.write(outdir, overwrite=True)

    # This will go as part of the writter
    with open(outdir / 'slurm.job', 'w') as f:
        f.write(SIMPLE_SLURM_DRIVER)



import sys
import logging
from pyschism.cmd import argument_parser
from pyschism.cmd.basecmd import SchismBaseCommand


class TidalRunCommand(SchismBaseCommand):
    """  Wrapper for SCHISM tidal only run CLI tool.  """


def main():
    parser = argument_parser.get_parser(
        'tidal',
        description="SCHISM tidal only run."
        )
    args = parser.parse_args()
    if len(args.constituents) == 0:
        args.constituents = ['all']
    logging.basicConfig(level=args.log_level)
    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    logging.getLogger("paramiko").setLevel(logging.WARNING)
    drv = TidalRunCommand(args)
    exit(drv.run())


# https://medium.com/opsops/how-to-test-if-name-main-1928367290cb
def init():
    if __name__ == "__main__":
        sys.exit(main())


init()



from datetime import datetime, timedelta
import os
import pathlib
import subprocess
from typing import Union

import numpy as np

from pyschism.domain import ModelDomain
from pyschism.driver.makefile import MakefileDriver
from pyschism.enums import Stratification
from pyschism.forcing.tides.bctides import Bctides
from pyschism.forcing.atmosphere import NWS2
from pyschism.param import Param
from pyschism.server import ServerConfig
from pyschism.stations import Stations


class CombineHotstartBinary:

    def __init__(self, path: Union[str, os.PathLike], iteration=None):
        path = pathlib.Path(path)
        if iteration is None:
            combine_hotstart = path.glob('hotstart_[0-9][0-9][0-9][0-9]_*.nc')
            increments = set([file.name.split('_')[-1].split('.nc')[0]
                              for file in combine_hotstart])
            iteration = np.max([int(increment) for increment in increments])
        subprocess.check_call(
            ["combine_hotstart7", '-i', f'{iteration}'], cwd=path)
        self.path = path / f"hotstart_it={iteration}.nc"


class ModelDriver:

    def __init__(
            self,
            model_domain: ModelDomain,
            dt: Union[float, timedelta],
            rnday: Union[float, timedelta],
            ihfskip: int = None,
            dramp: Union[float, timedelta] = None,
            start_date: datetime = None,
            ibc: Union[Stratification, int, str] = Stratification.BAROTROPIC,
            drampbc: Union[float, timedelta] = None,
            stations: Stations = None,
            nspool: Union[int, timedelta] = None,
            nhot_write: Union[int, timedelta, bool] = None,
            server_config: ServerConfig = None,
            combine_hotstart: Union[str, os.PathLike] = None,
            cutoff_depth: float = 50.,
            **surface_outputs):
        """Main driver class used to generate SCHISM input files

        This __init__ calls initialization routines for all SCHISM data inputs
        and prepares them to be dumped to files by the write() method.

        Arguments:
            mesh: :class:`pyschism.domain.ModelDomain`
            param: :class:`pyschism.param.param.Param`
        """
        # init param.nml
        self._param = Param(model_domain, dt, rnday, dramp, start_date, ibc,
                            drampbc, nspool, ihfskip, nhot_write, stations,
                            **surface_outputs)

        # init bctides.in
        self._bctides = Bctides(model_domain, self.param,
                                cutoff_depth=cutoff_depth)

        # init atmospheric data.
        self._nws = model_domain.nws

        # init hydrology data
        self._hydrology = model_domain.hydrology

        # init hotstart file
        self._combine_hotstart = combine_hotstart

        # init Makefile drivers
        self._makefile = MakefileDriver(server_config=server_config)

    def write(
            self,
            output_directory,
            overwrite=False,
            hgrid=True,
            vgrid=True,
            fgrid=True,
            param=True,
            bctides=True,
            nws=True,
            wind_rot=True,
            stations=True,
    ):
        """Writes to disk the full set of input files necessary to run SCHISM.
        """
        outdir = pathlib.Path(output_directory)

        if not (outdir / 'outputs').exists():
            (outdir / 'outputs').mkdir(parents=True)

        if hgrid:
            hgrid = 'hgrid.gr3' if hgrid is True else hgrid
            self.model_domain.hgrid.write(outdir / hgrid, overwrite)
            if self.model_domain.ics == 2:
                original_dir = os.getcwd()
                os.chdir(outdir)  # pushd
                try:
                    os.remove('hgrid.ll')
                except OSError:
                    pass
                os.symlink(hgrid, 'hgrid.ll')
                os.chdir(original_dir)  # popd
        if vgrid:
            vgrid = 'vgrid.in' if vgrid is True else vgrid
            self.model_domain.vgrid.write(outdir / vgrid, overwrite)
        if fgrid:
            fgrid = f'{self.model_domain.fgrid.fname}' if fgrid is True \
                    else fgrid
            self.model_domain.fgrid.write(outdir / fgrid, overwrite)
        if param:
            param = 'param.nml' if param is True else param
            self.param.write(outdir / param, overwrite)
        if bctides:
            bctides = 'bctides.in' if bctides is True else bctides
            self.bctides.write(outdir / bctides, overwrite)
        if nws:
            if self.nws is not None:
                if isinstance(self.nws, NWS2):
                    self.nws.write(outdir / 'sflux', overwrite,
                                   wind_rot=wind_rot)
                else:
                    self.nws.write(outdir, overwrite)
        if stations:
            if self.stations is not None:
                stations = 'station.in' if stations is True else stations
                self.stations.write(outdir / stations, overwrite)

        for hydrology in self._hydrology:
            hydrology.write(outdir, overwrite)

        if self.hotstart_file is not None:
            hotstart_lnk = outdir / 'hotstart.nc'
            if overwrite is True:
                if hotstart_lnk.exists():
                    hotstart_lnk.unlink()
            os.symlink(os.path.relpath(
                self.hotstart_file, outdir), hotstart_lnk)

        self.makefile.write(outdir / 'Makefile', overwrite)

    @property
    def param(self):
        return self._param

    @property
    def model_domain(self):
        return self.param.model_domain

    @property
    def stations(self):
        return self.param.stations

    @property
    def bctides(self):
        return self._bctides

    @property
    def nws(self):
        return self._nws

    @property
    def hydrology(self):
        return self._hydrology

    @property
    def hotstart_file(self):
        if self._combine_hotstart is not None:
            return self._combine_hotstart.path

    @property
    def makefile(self):
        return self._makefile

    @property
    def _nws(self):
        return self.__nws

    @_nws.setter
    def _nws(self, nws):
        if nws is not None:
            nws(self)
        self.__nws = nws

    @property
    def _combine_hotstart(self):
        return self.__combine_hotstart

    @_combine_hotstart.setter
    def _combine_hotstart(self, combine_hotstart):
        if combine_hotstart is not None:
            combine_hotstart = CombineHotstartBinary(combine_hotstart)
            self.param.opt.ihot = 1
        self.__combine_hotstart = combine_hotstart

    @property
    def _hydrology(self):
        return self.__hydrology

    @_hydrology.setter
    def _hydrology(self, hydrology):
        for forcing in hydrology:
            if callable(forcing):
                forcing(self)
        self.__hydrology = hydrology



from argparse import Namespace
from enum import Enum
import logging
import pathlib

from psutil import cpu_count

from pyschism.cmd.forecast.init import ForecastInit
from pyschism.cmd.forecast.update import ForecastUpdate
from pyschism.enums import ForecastProduct


class Dispatch(Enum):
    INIT = ForecastInit
    UPDATE = ForecastUpdate


class Env(Enum):
    INIT = 'init'
    UPDATE = 'update'


class ForecastCli:

    def __init__(self, args):
        Dispatch[Env(args.action).name].value(args)
        if args.action == "init":
            ForecastUpdate(Namespace(
                project_directory=args.project_directory,
                # log_level=args.log_level
            )
            )


def add_forecast_init(actions):
    init = actions.add_parser("init")
    init.add_argument("project_directory")
    init.add_argument('hgrid', help='Horizontal grid file.')
    init.add_argument('fgrid', help='Friction grid file.')
    init.add_argument('vgrid', nargs='?', help='Vertical grid file.')
    mesh_options = init.add_argument_group('mesh_options')
    mesh_options.add_argument('--hgrid-crs')
    mesh_options.add_argument('--fgrid-crs')
    mesh_options.add_argument(
        '--fgrid-type',
        choices=['auto', 'manning', 'drag', 'rough'],
        default='auto')
    init.add_argument("--timestep", type=float, required=True)
    init.add_argument("--forecast-days", type=float, required=True)
    init.add_argument(
        "--spinup-days",
        help="Number of days used for model initialization. "
        "Defaults to 15 days spinup.",
        type=float, default=15.)
    init.add_argument("--forecast-interval", type=float, default=24)
    init.add_argument("--timezone", default='UTC')
    init.add_argument(
        "--skip-run", action="store_true",
        help="Skips running the model.")
    init.add_argument('--nproc', type=int, default=cpu_count(logical=False))
    _add_tidal_constituents(init)
    _add_atmospheric_forcing(init)
    _add_hydrologic_forcing(init)
    # TODO: Additional forcings.
    # _add_wave_forcing(forecast)
    model_outputs = init.add_argument_group('model_outputs')
    _add_surface_outputs(model_outputs)
    # TODO: Stations outputs.
    # _add_stations_outputs(model_outputs)
    server_config = init.add_subparsers(dest="server_config")
    slurm = server_config.add_parser(
        'slurm', help="Add options for slurm run configuration.")
    slurm.add_argument('--account')
    slurm.add_argument('--partition')
    slurm.add_argument('--walltime', type=float, help="In hours, float.")
    slurm.add_argument('--slurm-filename')
    slurm.add_argument('--slurm-rundir')
    slurm.add_argument('--run-name')
    slurm.add_argument('--mail-type')
    slurm.add_argument('--mail-user')
    slurm.add_argument('--log-filename')
    slurm.add_argument('--path-prefix')
    slurm.add_argument('--slurm-nodes')
    slurm.add_argument('--slurm-launcher', default='srun')
    slurm.add_argument('--extra-commands', action='append')
    slurm.add_argument(
        '--module',
        default=list(),
        action='append',
        dest='modules')


def add_forecast_update(actions):
    update = actions.add_parser("update")
    update.add_argument("project_directory")


def add_forecast(subparsers):
    forecast = subparsers.add_parser('forecast')
    forecast.add_argument(
        "--overwrite", action="store_true",
        help="Allow overwrite of output directory.")
    forecast.add_argument(
        "--log-level",
        choices=[name.lower() for name in logging._nameToLevel])
    actions = forecast.add_subparsers(dest="action")
    add_forecast_init(actions)
    add_forecast_update(actions)


def add_forecastd(subparsers):
    forecastd = subparsers.add_parser('forecastd')
    actions = forecastd.add_subparsers(dest='action')
    actions.required = True
    actions.add_parser('start')
    actions.add_parser('stop')
    actions.add_parser('restart')
    add = actions.add_parser('add')
    add.add_argument("-o", "--output-directory", dest='outdir')
    _add_mesh_options(add)
    _add_tidal_constituents(add)
    _add_atmospheric_forcing(add)
    _add_hydrologic_forcing(add)


def add_viewerd(subparsers):
    viewerd = subparsers.add_parser('viewerd')
    actions = viewerd.add_subparsers(dest='action')
    actions.required = True
    start = actions.add_parser('start')
    start.add_argument('--deploy', action='store_true')
    actions.add_parser('stop')
    actions.add_parser('restart')


def add_autodocd(subparsers):
    autodocd = subparsers.add_parser('autodocd')
    actions = autodocd.add_subparsers(dest='action')
    actions.required = True
    actions.add_parser('start')
    actions.add_parser('stop')
    actions.add_parser('restart')


def add_plot(subparsers):
    plot = subparsers.add_parser('plot')
    plot.add_argument(
        'resource', type=pathlib.Path,
        help='Filename or directory containing SCHISM outputs.')
    plot.add_argument('variable', help='Name of variable to plot.')
    output_type = plot.add_subparsers(dest='output_type')
    output_type.required = True
    surface_plot = output_type.add_parser('surface', help="Plot SCHISM surface"
                                          " outputs.")
    surface_plot.add_argument("--start")
    stations = output_type.add_parser('stations', help="Plot SCHISM stations "
                                      "outputs.")
    stations.add_argument('-s', '--station', nargs='*', type=int,
                          help='Station index in station.in to include in '
                          'plot.')


def _add_mesh_options(parser):
    parser.add_argument('hgrid', help='Horizontal grid file.')
    parser.add_argument('fgrid', help='Friction grid file.')
    mesh_options = parser.add_argument_group('mesh_options')
    mesh_options.add_argument('--vgrid', help='Vertical grid file.')
    mesh_options.add_argument('--hgrid-crs')
    mesh_options.add_argument('--fgrid-crs')


def _add_tidal_constituents(parser):
    tides = parser.add_argument_group('tides')
    options = tides.add_mutually_exclusive_group()
    options.required = True
    options.add_argument("--all-constituents", action="store_true")
    options.add_argument("--major-constituents", action="store_true")
    options.add_argument(
        "-c", "--constituents",
        action='append',
        choices=["K1", "O1", "P1", "Q1", "MM", "Mf", "M4", "MN4", "MS4",
                 "2N2", "S1"],
        dest='constituents',
        default=False,
        help="Tidal constituent to be forced in the model (case-insensitive).")
    tides.add_argument("--include-tidal-velocity",
                       "--bnd-vel", action="store_true", dest='bnd_vel')


def _add_atmospheric_forcing(parser):
    atmospheric_forcing_1 = parser.add_argument_group(
        'atmospheric forcing level 1')
    data_source_1 = atmospheric_forcing_1.add_mutually_exclusive_group()
    for product in ForecastProduct:
        data_source_1.add_argument(
            f"--{product.value.lower().replace('_', '-')}",
            action="store_true",
            help=f'Use {product.value} as data source.')
    data_source_1.add_argument(
        "--gfs", action="store_true",
        help="Alias for --gfs-0p25-1hr")
    data_source_1.add_argument(
        "--gdas", action='store_true',
        help="Alias for --gdas-0p25",
    )
    atmospheric_forcing_1.add_argument(
        "--no-air-1", action="store_true",
        help="Disables air mass forcing option.")
    atmospheric_forcing_1.add_argument(
        "--no-prc-1", action='store_true',
        help="Disables precipitation option.")
    atmospheric_forcing_1.add_argument(
        "--no-rad-1", action='store_true',
        help="Disables solar radiation flux option.")

    atmospheric_forcing_2 = parser.add_argument_group(
        'atmospheric forcing level 2')
    # data_source_2 = atmospheric_forcing_2.add_mutually_exclusive_group()
    # for product in ["HWRF", "HRRR", "ETC"]:
    #     data_source_2.add_argument(f"--{product}", help=f'Use {product} as '
    #                                'data source.')
    atmospheric_forcing_2.add_argument(
        "--no-air-2", action="store_true",
        help="Disables air mass forcing option.")
    atmospheric_forcing_2.add_argument(
        "--no-prc-2", action='store_true',
        help="Disables precipitation option.")
    atmospheric_forcing_2.add_argument(
        "--no-rad-2", action='store_true',
        help="Disables solar radiation flux option.")


def _add_hydrologic_forcing(parser):
    src_snk = parser.add_argument_group('Sources and sinks')
    src_snk.add_argument("--hydrology", action="append",
                         help="Add source and sink flow.",
                         choices=['NWM'],
                         default=[])


def _add_wave_forcing(parser):
    wave_forcing = parser.add_argument_group('Waves')
    data_source = wave_forcing.add_mutually_exclusive_group()
    for product in ["WWIII"]:
        data_source.add_argument(f'--{product}')


def hydro():
    """
    hydro output options
    """
    return {
        1: ("elev", "0: off; 1: on - elev. [m]"),
        2: ("air_pressure", "air pressure [Pa]"),
        3: ("air_temperature", "air temperature [C]"),
        4: ("specific_humidity", "Specific humidity [-]"),
        5: ("solar_radiation", "solar (shortwave) radiation [W/m/m]"),
        6: ("sensible_flux", "sensible flux (positive upward) [W/m/m]"),
        7: ("latent_heat", "latent heat flux (positive upward) [W/m/m]"),
        8: ("upward_longwave",
            "upward longwave radiation (positive upward) [W/m/m]"),
        9: ("downward_longwave",
            "downward longwave radiation (positive downward) [W/m/m]"),
        10: ("total_heat_flux", "total flux=-flsu-fllu-(radu-radd) [W/m/m]"),
        11: ("evaporation", "evaporation rate [kg/m/m/s]"),
        12: ("precipitation", "precipitation rate [kg/m/m/s]"),
        13: ("bottom_stress", "Bottom stress vector [kg/m/s^2(Pa)]"),
        14: ("wind_speed", "wind velocity vector [m/s]"),
        15: ("wind_stress", "wind stress vector [m^2/s/s]"),
        16: ("dahv", "depth-averaged vel vector [m/s]"),
        17: ("vertical_velocity", "vertical velocity [m/s]"),
        18: ("temp", "water temperature [C]"),
        19: ("salt", "water salinity [PSU]"),
        20: ("water_density", "water density [kg/m^3]"),
        21: ("diffusivity", "eddy diffusivity [m^2/s]"),
        22: ("viscosity", "eddy viscosity [m^2/s]"),
        23: ("TKE", "turbulent kinetic energy"),
        24: ("mixing-lenght", "turbulent mixing length [m]"),
        25: ("hvel", "horizontal vel vector [m/s]"),
        26: ("hvel_side", "horizontal vel vector defined @side [m/s]"),
        27: ("wvel_elem", "vertical vel. @elem [m/s]"),
        28: ("temp_elem", "T @prism centers [C]"),
        29: ("salt_elem", "S @prism centers [PSU]"),
        30: ("pressure_gradient",
             "Barotropic pressure gradient force vector (m.s-2) @side "
             "centers"),

    }


def wwm():
    """
    WWM output options
    """
    return {
        1: ("WWM_1", "sig. height (m)"),
        2: ("WWM_2", "Mean average period (sec) - TM01"),
        3: ("WWM_3",
            "Zero down crossing period for comparison with buoy (s) - TM02"),
        4: ("WWM_4", "Average period of wave runup/overtopping - TM10"),
        5: ("WWM_5", "Mean wave number (1/m)"),
        6: ("WWM_6", "Mean wave length (m)"),
        7: ("WWM_9",
            "Mean average energy transport direction (degr) - MWD in NDBC?"),
        8: ("WWM_10", "Mean directional spreading (degr)"),
        9: ("WWM_11", "Discrete peak period (sec) - Tp"),
        10: ("WWM_12",
             "Continuous peak period based on higher order moments (sec)"),
        11: ("WWM_13", "Peak phase vel. (m/s)"),
        12: ("WWM_14", "Peak n-factor."),
        13: ("WWM_15", "Peak group vel. (m/s)"),
        14: ("WWM_16", "Peak wave number"),
        15: ("WWM_17", "Peak wave length"),
        16: ("WWM_18", "Peak (dominant) direction (degr)"),
        17: ("WWM_19", "Peak directional spreading"),
        18: ("WWM_20", "Discrete peak direction (radian?) "),
        19: ("WWM_21", "Orbital vel. (m/s) "),
        20: ("WWM_22", "RMS Orbital vel. (m/s) "),
        21: ("WWM_23", "Bottom excursion period (sec?) "),
        22: ("WWM_24", "Bottom wave period (sec) "),
        23: ("WWM_25", "Uresell number based on peak period "),
        24: ("WWM_26", "Friction velocity (m/s?) "),
        25: ("WWM_27", "Charnock coefficient "),
        26: ("WWM_28", "Rougness length "),
        27: ("WWM_energy_dir", "WWM_energy vector"),
        28: ("wave-force",
             "Wave force vector (m.s-2) computed by wwm @side centers and "
             "whole levels"),
    }


def gen():
    """
    gen output options
    """
    return {
        1: ("GEN_1", "1st tracer"),
        2: ("GEN_2", "2nd tracer"),
    }


def age():
    """
    age output options
    """
    return {
        1: ("AGE_1", "Indices from \"1\" to \"ntracer_age/2\"; [days]"),
        2: ("AGE_2", "Indices from \"1\" to \"ntracer_age/2\"; [days]"),
    }


def sed():
    """
    sed output options
    """
    return {
        1: ("SED_depth_change",
            "bottom depth _change_ from init. condition (m)"),
        2: ("SED_D50", " Bed median grain size in the active layer (mm)"),
        3: ("SED_bed_stress", " Bottom shear stress (Pa)"),
        4: ("SED_bed_roughness", " Bottom roughness lenghth (mm)"),
        5: ("SED_TSC", "total suspended concentration (g/L)"),
        6: ("bed_thickness", " total bed thickness @elem (m)"),
        7: ("bed_age", " total bed age over all layers @elem (sec)"),
        8: ("z0st",
            " Sediment transport roughness length @elem (m) (z0st_elem)"),
        9: ("z0cr", "current-ripples roughness length @elem (m) (z0cr_elem)"),
        10: ("z0sw", "sand-waves roughness length (m) @elem (z0sw_elem)"),
        11: ("z0wr", "wave-ripples roughness length @elem (m) (z0wr_elem)"),
        12: ("SED3D_1",
             "conc. of 1st class (one output need by each class) [g/L]"),
        13: ("SED_bdld_1",
             "Bedload transport rate vector (kg.m-1.s-1) for 1st tracer (one "
             "output need by tracer)"),
        14: ("SED_bedfrac_1",
             "Bed fraction 1st tracer (one output need by each class) [-]"),
        15: ("SED3D_2", "conc. of 2nd class"),
        16: ("SED_bdld_2", "Bedload transport of 2nd class"),
        17: ("SED_bedfrac_3", "Bed fraction of 2nd class"),
    }


def eco():
    """
    EcoSim output options
    """
    return {
        1: ("ECO_1", "EcoSim outputs")
    }


def icm():
    """
    ICM output options
    """
    return {
        1: ("ICM_Chl", "Chlorophyll"),
        2: ("ICM_pH", "PH values (ICM_PH on)"),
        3: ("ICM_PrmPrdt", "ICM primary production @elem [gC/m^3/day]"),
        4: ("ICM_DIN", "ICM totoal inorganic nitrogen (DIN) @elem [gN/m^3]"),
        5: ("ICM_PON", "ICM paticulate organic nitrogen (PON) @elem [gN/m^3]"),
        6: ("ICM_SED_BENDOC",
            "ICM bed sediment flux arrays: SED_BENDOC (output "
            "name:ICM_SED_BENDOC) @elem [gC/(m^2 day)]"),
        7: ("ICM_SED_BENNH4",
            "ICM bed sediment flux arrays: SED_BENNH4 (output "
            "name:ICM_SED_BENNH4) @elem [gC/(m^2 day)]"),
        8: ("ICM_SED_BENNO3",
            "ICM bed sediment flux arrays: SED_BENNO3 (output "
            "name:ICM_SED_BENNO3)@elem [gC/(m^2 day)]"),
        9: ("ICM_SED_BENPO4",
            "ICM bed sediment flux arrays: SED_BENPO4 (output "
            "name:ICM_SED_BENPO4) @elem [gC/(m^2 day)]"),
        10: ("ICM_SED_BENCOD",
             "ICM bed sediment flux arrays: SED_BENCOD (output "
             "name:ICM_SED_BENCOD) @elem [gC/(m^2 day)]"),
        11: ("ICM_SED_BENDO",
             "ICM bed sediment flux arrays: SED_BENDO (output "
             "name:ICM_SED_BENDO) @elem [gC/(m^2 day)]"),
        12: ("ICM_SED_BENSA",
             "ICM bed sediment flux arrays: SED_BENSA (output "
             "name:ICM_SED_BENSA) @elem [gC/(m^2 day)]"),
        13: ("ICM_lfsav",
             "ICM SAV leaf biomass @elem [gC/m^3] (k=1 is surface)"),
        14: ("ICM_stsav", "ICM SAV stem biomass @elem [gC/m^3]"),
        15: ("ICM_rtsav", "ICM SAV root biomass @elem [gC/m^3]"),
        16: ("ICM_tlfsav", "ICM SAV total leaf biomass @elem [gC/m^2]"),
        17: ("ICM_tstsav", "ICM SAV total stem biomass @elem [gC/m^2]"),
        18: ("ICM_trtsav", "ICM SAV total root biomass @elem [gC/m^2]"),
        19: ("ICM_hcansav", "ICM SAV canopy height @elem [m]"),
        20: ("ICM_CNH4", "bottom NH4 conc"),
        21: ("ICM_CNH3", "bottom NO3 conc"),
        22: ("ICM_CPIP", "bottom P conc"),
        23: ("ICM_CPOS", "bottom Si conc"),
        24: ("ICM_CCH4", "bottom CH4 conc"),
        25: ("ICM_CSO4", "bottom SO4 conc"),
        26: ("ICM_CH2S", "bottom H2S conc"),
        27: ("ICM_SEDPON1", "bottom PON g1 conc"),
        28: ("ICM_SEDPON2", "bottom PON g2 conc"),
        29: ("ICM_SEDPON3", "bottom PON g3 conc"),
        30: ("ICM_SEDPOP1", "bottom POP g1 conc"),
        31: ("ICM_SEDPOP2", "bottom POP g2 conc"),
        32: ("ICM_SEDPOP3", "bottom POP g3 conc"),
        33: ("ICM_SEDPOC1", "bottom POC g1 conc"),
        34: ("ICM_SEDPOC2", "bottom POC g2 conc"),
        35: ("ICM_SEDPOC3", "bottom POC g3 conc"),
        36: ("ICM_EROH2S", "erosion flux H2S"),
        37: ("ICM_EROLPOC", "ersoion flux LPOC"),
        38: ("ICM_ERORPOC", "ersoion flux RPOC"),
        39: ("ICM_DO_consumption", "DO consumption"),
        40: ("ICM_GP1", "PB growth #1"),
        41: ("ICM_GP2", "PB growth #2"),
        42: ("ICM_GP3", "PB growth #3"),
        43: ("ICM_1", "Zoo. #1"),
        44: ("ICM_2", "Zoo. #2"),
        45: ("ICM_3", "phyto #1"),
        46: ("ICM_4", "phyto #2"),
        47: ("ICM_5", "phyto #3"),
        48: ("ICM_6", "RPOC"),
        49: ("ICM_7", "LPOC"),
        50: ("ICM_8", "DOC"),
        51: ("ICM_9", "RPON"),
        52: ("ICM_10", "LPON"),
        53: ("ICM_11", "DON"),
        54: ("ICM_12", "NH4"),
        55: ("ICM_13", "NO3"),
        56: ("ICM_14", "RPOP"),
        57: ("ICM_15", "LPOP"),
        58: ("ICM_16", "DOP"),
        59: ("ICM_17", "PO4t"),
        60: ("ICM_18", "Si- biogenic"),
        61: ("ICM_19", "available Si"),
        62: ("ICM_20", "COD: Chemical oxygen demand"),
        63: ("ICM_21", "DO"),
        64: ("ICM_22", "TIC"),
        65: ("ICM_23", "ALK"),
        66: ("ICM_24", "CA"),
        67: ("ICM_25", "CACO3"),
    }


def cos():
    """
    CoSINE output options
    """
    return {
        1: ("COS_1", "COS_1"),
        2: ("COS_2", "COS_2"),
        3: ("COS_3", "COS_3"),
        4: ("COS_4", "COS_4"),
        5: ("COS_5", "COS_5"),
        6: ("COS_6", "COS_6"),
        7: ("COS_7", "COS_7"),
        8: ("COS_8", "COS_8"),
        9: ("COS_9", "COS_9"),
        10: ("COS_10", "COS_10"),
        11: ("COS_11", "COS_11"),
        12: ("COS_12", "COS_12"),
        13: ("COS_13", "COS_13"),
    }


def fib():
    """
    Fecal indicating bacteria output options
    """
    return {
        1: ("FIB_1", "FIB_1")
    }


def sed2d():
    """
    SED2D output options
    """
    return {
        1: ("SED2D_depth_change",
            "bottom depth _change_ from init. condition (m)"),
        2: ("SED2D_Cd", "drag coefficient used in transport formulae"),
        3: ("SED2D_cflsed", "Courant number (b.qtot.dt / h.dx)"),
        4: ("SED2D_d50", "Top layer d50 (m)"),
        5: ("SED2D_total_transport", "total transport rate vector (kg/m/s)"),
        6: ("SED2D_susp_load", "suspended tranport rate vector (kg/m/s)"),
        7: ("SED2D_bed_load", "bedload transport rate vector (kg/m/s)"),
        8: ("SED2D_average_transport",
            "time averaged total transport rate vector (kg/m/s)"),
        9: ("SED2D_bottom_slope",
            "bottom slope vector (m/m); negative uphill"),
        10: ("z0eq2d", "Total roughness length @elem (m) (z0eq)"),
        11: ("z0cr2d", "current-ripples roughness length @elem (m) (z0cr)"),
        12: ("z0sw2d", "sand-waves roughness length @elem (m) (z0sw)"),
        13: ("z0wr2d", "wave-ripples roughness length @elem (m) (z0wr)"),
    }


def mar():
    """
    marsh output options
    """
    return {
        1: ("marsh_flag", "marsh_flag"),
    }


def ice():
    """
    ice output options
    """
    return {
        1: ("ICE_velocity", "ice advective velcoity vector [m/s]"),
        2: ("ICE_strain_rate", "strain rate @ elem [1/sec]"),
        3: ("ICE_net_heat_flux",
            "net heat flux to ocean (>0 warm up SST) [W/m/m]"),
        4: ("ICE_fresh_water_flux",
            "net fresh water flux to ocean (>0 freshens up SSS) [kg/s/m/m]"),
        5: ("ICE_top_T", "ice temperature [C] at air-ice interface"),
        6: ("ICE_tracer_1", "ice volume [m]"),
        7: ("ICE_tracer_2", "ice concentration [-]"),
        8: ("ICE_tracer_3", "snow volume [m]"),
    }


def ana():
    return {
        1: ("ANA_air_pres_grad_x", "x-component of 𝛁air_pres/ρ0 [m/s/s]"),
        2: ("ANA_air_pres_grad_y", "y-component of 𝛁air_pres/ρ0 [m/s/s]"),
        3: ("ANA_tide_pot_grad_x",
            "α*g*𝛁Ψ [m/s/s] (gradient of tidal potential)"),
        4: ("ANA_tide_pot_grad_y", "α*g*𝛁Ψ [m/s/s]"),
        5: ("ANA_hor_viscosity_x", "𝛁·(μ𝛁u) [m/s/s] (horizontal viscosity)"),
        6: ("ANA_hor_viscosity_y", "𝛁·(μ𝛁u) [m/s/s]"),
        7: ("ANA_bclinic_force_x",
            "-g/rho0* ∫_z^η dr_dx dz  [m/s/s] (b-clinic gradient)"),
        8: ("ANA_bclinic_force_y", "-g/rho0* ∫_z^η dr_dy dz  [m/s/s]"),
        9: ("ANA_vert_viscosity_x",
            "d (ν du/dz)/dz [m/s/s] - no vegetation effects (vertical "
            "viscosity)"),
        10: ("ANA_vert_viscosity_y",
             "d (ν dv/dz)/dz [m/s/s] - no vegetation effects"),
        11: ("ANA_mom_advection_x", "(u·𝛁) u [m/s/s] (momentum advection)"),
        12: ("ANA_mom_advection_y", "(u·𝛁) u [m/s/s]"),
        13: ("ANA_Richardson", "gradient Richardson number [-]"),
        14: ("ANA_transport_min_dt_elem",
             "min time step at each element over all subcycles in horizontal "
             "transport solver [s]  "),
    }


def _add_surface_outputs(parser):

    parser.add_argument(
        '--nspool',
        help='If passing an integer, it is interpreted as timestep, '
             'if passing a float, it is interpreted as hours.')

    outputs = {
        'hyd': hydro(),
        'wwm': wwm(),
        'gen': gen(),
        'age': age(),
        'sed': sed(),
        'eco': eco(),
        'icm': icm(),
        'cos': cos(),
        'fib': fib(),
        'sed2d': sed2d(),
        'mar': mar(),
        'ice': ice(),
        'ana': ana(),
    }
    for short_name, output in outputs.items():
        for id, (long_name, help_msg) in output.items():
            parser.add_argument(
                f"--{long_name.lower().replace('_', '-')}",
                f"-{short_name}{id}",
                help=help_msg,
                action='store_true'
            )



from datetime import datetime, timedelta
import logging
import pathlib
from typing import Union

from pyschism.domain import ModelDomain
from pyschism.enums import Stratification
from pyschism.param.core import CORE
from pyschism.param.opt import OPT
from pyschism.param.schout import SCHOUT
from pyschism.stations import Stations

_logger = logging.getLogger(__name__)

class Param:

    def __init__(
            self,
            model_domain: ModelDomain,
            dt: Union[int, float, timedelta],
            rnday: Union[int, float, timedelta],
            dramp: Union[int, float, timedelta] = None,
            start_date: datetime = None,
            ibc: Union[Stratification, int, str] = Stratification.BAROTROPIC,
            drampbc: Union[int, float, timedelta] = None,
            nspool: Union[int, float, timedelta] = None,
            ihfskip: Union[int, timedelta] = None,
            nhot_write: Union[int, timedelta, bool] = None,
            stations: Stations = None,
            **surface_outputs):

        _logger.info('Initializing param')
        self._model_domain = model_domain
        self._core = CORE(ibc, rnday, dt, nspool, ihfskip)
        self._opt = OPT(dramp, drampbc, start_date)
        self._schout = SCHOUT(**surface_outputs)
        self._nhot_write = nhot_write
        self._stations = stations

    def __str__(self):
        return f"{str(self.core)}\n\n{str(self.opt)}\n\n{str(self.schout)}\n"

    def write(self, path, overwrite=False, use_template=False):
        path = pathlib.Path(path)
        if path.is_file() and not overwrite:
            raise IOError(f"File {path} exists and overwrite=False")
        with open(path, 'w') as f:
            f.write(str(self))

    @property
    def core(self):
        return self._core

    @property
    def opt(self):
        return self._opt

    @property
    def schout(self):
        return self._schout

    @property
    def model_domain(self):
        return self.__model_domain

    @property
    def nhot_write(self):
        return self.schout.nhot_write

    @property
    def stations(self):
        return self.__stations

    @property
    def _model_domain(self):
        return self.__model_domain

    @_model_domain.setter
    def _model_domain(self, model_domain):
        assert isinstance(model_domain, ModelDomain), \
            f"Argument model_domain must be of type {ModelDomain}, " \
            f"not {type(model_domain)}."
        self.__model_domain = model_domain

    @property
    def _opt(self):
        return self.__opt

    @_opt.setter
    def _opt(self, opt: OPT):
        # friction parameters
        opt.nchi = self.model_domain.fgrid
        # set coordinate system
        opt.ics = self.model_domain.ics
        # set coriolis
        opt.ncor = self.model_domain.ncor
        # set atmospheric forcing
        if self.model_domain.nws is not None:
            opt.nws = self.model_domain.nws
        # TODO: Set the remaining options:
        # msc2
        # mdc2
        # ntracer_gen
        # ntracer_age
        # sed_class
        # eco_class
        self.__opt = opt

    @property
    def _nhot_write(self):
        return self.__nhot_write

    @_nhot_write.setter
    def _nhot_write(self, nhot_write: Union[int, timedelta, bool, None]):

        if not isinstance(nhot_write, (int, bool, timedelta, type(None))):
            raise TypeError(f"Argument nhot_write must be of type {int}, "
                            f"{bool}, {timedelta}, or None.")

        if nhot_write is True:
            nhot_write = int(round(self.core.rnday / self.core.dt))

        elif isinstance(nhot_write, timedelta):
            nhot_write = int(round(nhot_write / self.core.dt))

        if nhot_write is not None:
            if nhot_write % self.core.ihfskip != 0:
                raise ValueError("nhot_write must be a multiple of ihfskip")
            self.schout.nhot_write = nhot_write

    @property
    def _stations(self):
        return self.__stations

    @_stations.setter
    def _stations(self, stations: Union[Stations, None]):
        assert isinstance(stations, (Stations, type(None))), \
            f"Argument stations must be of type {Stations} or None, " \
            f"not type {type(stations)}."
        if isinstance(stations, Stations):
            raise NotImplementedError("Do something!")
        self.__stations = stations



#! /usr/bin/env python

from pyschism.outputs import (
    PlotOutputCombined,
    PlotOutputStations,
    StationsOutput,
)


class OutputPlotCli:

    def __init__(self, args):

        if not args.resource.exists():
            raise OSError(f'ERROR: Invalid path {str(args.resource)}. '
                          'It does not exist.')

        self._args = args

        # args.output_type is at current a required argument but in the future
        # it doesn't have to be required.
        if args.output_type is None:
            # will never hit this part unless we remove the output_type req.
            raise NotImplementedError('Should plot everything.')

        else:
            getattr(self, args.output_type)()

    def surface(self):

        if self._args.resource.is_dir():
            raise NotImplementedError(
                'Uncombined outputs not yet supported, please pass a combined '
                'output and stay tuned for updates. This will be supported in '
                'the future.')

        elif self._args.resource.is_file():
            plotter = PlotOutputCombined(self._args.resource)
            plotter.plot(self._args.variable, show=True)

    def stations(self):
        plotter = PlotOutputStations(StationsOutput(self._args.resource))
        plotter.plot(self._args.variable)



#! /usr/bin/env python
import argparse
import pathlib
from functools import lru_cache
from scipy.interpolate import griddata, interp1d
import matplotlib.pyplot as plt
from pyproj import Transformer
from datetime import datetime
import scipy.io as sio
import numpy as np
from datetime import timedelta
import imageio
import os
from pyschism.mesh import Hgrid


wt = os.getenv('TCPROF_WIND_TYPE')
wt = 'GAHM' if wt is None else wt


PARENT = pathlib.Path(__file__).parent.resolve()
FILE = PARENT / f'TC_Prof_all_weighted_composite_R2_{wt}-v6.mat'


class TCProf:

    def __init__(self, event_id):
        self._event_id = event_id

    def interpolate(
        self,
        hgrid,
        output_path,
        timestep,
        nskip=1,
        overwrite=False
    ):

        # certify hgrid
        msg = f"hgrid must be of type {Hgrid}"
        assert isinstance(hgrid, Hgrid), msg
        assert isinstance(nskip, int)

        xy = hgrid.get_xy(crs="EPSG:4326")

        def interp_field(step, x, y, field):
            return griddata(
                (x(step), y(step)),
                field(step),
                (xy[:, 0], xy[:, 1]),
                method='linear',
                fill_value=0.
                )

        # certify output_path
        output_path = pathlib.Path(output_path)
        if output_path.is_file() and not overwrite:
            msg = f"File {output_path} exists and overwrite=False"
            raise IOError(msg)

        steps, x, y, w, d, p = self._interp_hgrid(hgrid, timestep, nskip)

        with open(output_path, 'w') as f:
            for step in steps:
                line = f"{step.total_seconds()}, "
                wdata = interp_field(step, x, y, w)
                ddata = interp_field(step, x, y, d)
                pdata = interp_field(step, x, y, p)
                for i in range(hgrid.values.size):
                    u = wdata[i]*np.cos(ddata[i])
                    v = wdata[i]*np.sin(ddata[i])
                    line += f"{u} {v} {pdata[i]}, "
                line = line[:-2] + '\n'
                f.write(line)

    def plot_wfield(self, iframe, dsfact=6, scale=5000):
        fig, ax = plt.subplots(figsize=(16, 12))
        values = self.wind_magnitude[iframe]
        values = np.ma.masked_where(values < 64., values)
        ax_wind = ax.quiver(
            self.lon_arrays[iframe][::dsfact, ::dsfact],
            self.lat_arrays[iframe][::dsfact, ::dsfact],
            self.u_arrays[iframe][::dsfact, ::dsfact],
            self.v_arrays[iframe][::dsfact, ::dsfact],
            values[::dsfact, ::dsfact],
            scale=scale,
            cmap='jet'
            )
        ax.axis('scaled')
        fig.suptitle('Interpolated Wind Field of TC: ' + self.event_id)
        ax.set_title(self.datetime[iframe].strftime('%b %d, %Y %H:%M'))
        ax.set_xlabel('Longitude (°E)')
        ax.set_ylabel('Latitude (°N)')
        cbar = plt.colorbar(ax_wind, format='%.1f')
        cbar.ax.set_ylabel('Wind speed [km/h]', rotation=90)
        plt.show()
        plt.close(plt.gcf())

    def plot_pfield(self, iframe):
        fig, ax = plt.subplots(figsize=(16, 12))
        # ax.plot(self.coast[:, 1], self.coast[:, 0], color='black',
        #         linewidth=0.5)
        values = self.pressure[iframe]
        values = np.ma.masked_where(self.wind_magnitude[iframe] < 64., values)
        ax_press = ax.contourf(
            self.lon_arrays[iframe],
            self.lat_arrays[iframe],
            values,
            levels=256,
            cmap='jet',
            # vmax=self.plot_vmax,
            # vmin=self.plot_vmin
            )
        ax.axis('scaled')
        fig.suptitle('Interpolated Pressure Field of TC: ' + self.event_id)
        ax.set_title(self.datetime[iframe].strftime('%b %d, %Y %H:%M'))
        ax.set_xlabel('Longitude (°E)')
        ax.set_ylabel('Latitude (°N)')
        cbar = plt.colorbar(ax_press, format='%.1f')
        cbar.ax.set_ylabel('Pressure [hPa]', rotation=90)
        plt.show()
        plt.close(plt.gcf())

    def make_gif(self):
        images = list()
        dirname = self.filename
        os.makedirs(dirname, exist_ok=True)
        for i, time in enumerate(self.datetime):
            fig, ax = plt.subplots(figsize=(16, 12))
            ax.plot(self.coast[:, 1], self.coast[:, 0], color='black',
                    linewidth=0.5)
            values = self.values[i]
            values = np.ma.masked_where(values < 64., values)
            ax_wind = ax.contourf(self.lon_arrays[i], self.lat_arrays[i],
                                  values, levels=512, cmap='jet',
                                  vmax=self.plot_vmax, vmin=self.plot_vmin
                                  )
            ax.axis('scaled')
            fig.suptitle('Interpolated Wind Field of TC: ' + self.name[i])
            ax.set_title(time.strftime('%b %d, %Y %H:%M'))
            ax.set_xlabel('Longitude (°E)')
            ax.set_ylabel('Latitude (°N)')
            cbar = plt.colorbar(ax_wind, format='%.1f')
            cbar.ax.set_ylabel('Wind speed km/h', rotation=90)

            plt.savefig('{}/{:d}.png'.format(dirname, i))
            images.append(imageio.imread('{}/{:d}.png'.format(dirname, i)))
            plt.close()
        imageio.mimsave('{}/full_movie.gif'.format(dirname), images, fps=2)

    @property
    @lru_cache
    def tcprof(self):
        tcprof = {}
        for i, lines in enumerate(sio.loadmat(FILE)['TC_Prof'][0, :]):
            name = lines[0][0, 0][6][0]
            year = lines[0][0, 0][8][0]
            storm = (name + year).capitalize()
            tcprof[storm] = {
                "wind": [],
                "press": [],
                "param": [],
                "radio": [],
                "datetime": [],
            }
            for entry in lines[0][0, :]:
                if np.any(np.isnan(entry[3])):
                    continue
                tcprof[storm]['wind'].append(entry[0])
                tcprof[storm]['press'].append(entry[1])
                tcprof[storm]['param'].append(entry[3])
                tcprof[storm]['radio'].append(entry[5])
                tcprof[storm]['datetime'].append(
                    datetime.strptime(
                        f"{entry[9][0]} {entry[8][0]}",
                        "%b %d %H:%M %Y")
                    )
            tcprof[storm]['storm_id'] = f'AL{int(entry[7]):02d}{year}'
            tcprof[storm]['matlab_id'] = f'{int(i)+1}'
        return tcprof

    @property
    def event_id(self):
        return self._event_id

    @property
    def datetime(self):
        return self.tcprof[self.event_id]['datetime']

    @property
    @lru_cache
    def wind_magnitude(self):
        wind_magnitude = list()
        for date in self.datetime:
            wind_magnitude.append(self.wfields[date])
        return wind_magnitude

    @property
    @lru_cache
    def pressure(self):
        press = list()
        for date in self.datetime:
            press.append(self.pfields[date])
        return press

    @property
    @lru_cache
    def wfields(self):
        r_out, theta_out = np.meshgrid(
                np.linspace(0, 500000, 501),
                np.linspace(0., 2.*np.pi, 501)
                )
        fields = {}
        for i, date in enumerate(self.datetime):
            w_input = list()
            for j, angle in enumerate(np.linspace(-np.pi/4, 9*np.pi/4, num=6)):
                w_input.append(
                    self.tcprof[self.event_id]['wind'][i][((j-1) % 4), :])
            w_input = np.array(w_input).flatten()
            wfield = griddata(
                (self.r_input, self.theta_input),
                w_input,
                (r_out.flatten(), theta_out.flatten()),
                method='linear')
            # wfield interp
            idx = np.where(np.isnan(wfield))[0]
            _idx = np.where(~np.isnan(w_input))[0]
            _wfield = griddata(
                (self.r_input[_idx], self.theta_input[_idx]),
                w_input[_idx],
                (r_out.flatten()[idx], theta_out.flatten()[idx]),
                method='nearest')
            for i, _idx in enumerate(idx):
                wfield[_idx] = _wfield[i]
            wfield = wfield.reshape(r_out.shape)
            fields[date] = wfield
        return fields

    @property
    @lru_cache
    def pfields(self):
        r_out, theta_out = np.meshgrid(
                np.linspace(0, 500000, 501),
                np.linspace(0., 2.*np.pi, 501)
                )
        fields = {}
        for i, date in enumerate(self.datetime):
            p_input = list()
            for j, angle in enumerate(np.linspace(-np.pi/4, 9*np.pi/4, num=6)):
                p_input.append(
                        self.tcprof[self.event_id]['press'][i][((j-1) % 4), :],
                        )
            p_input = np.array(p_input).flatten()
            pfield = griddata(
                (self.r_input, self.theta_input),
                p_input,
                (r_out.flatten(), theta_out.flatten()),
                method='linear')
            # pfield interp
            idx = np.where(np.isnan(pfield))
            if len(idx[0]) > 0:
                _idx = np.where(~np.isnan(p_input))
                _pfield = griddata(
                    (self.r_input[_idx], self.theta_input[_idx]),
                    p_input[_idx],
                    (r_out.flatten()[idx], theta_out.flatten()[idx]),
                    method='nearest')
                for i, _idx in enumerate(idx):
                    pfield[_idx] = _pfield[i]
            pfield = pfield.reshape(r_out.shape)
            fields[date] = pfield
        return fields

    @property
    @lru_cache
    def wdirfields(self):
        r_out, theta_out = np.meshgrid(
                np.linspace(0, 500000, 501),
                np.linspace(0., 2.*np.pi, 501)
                )
        fields = {}
        for i, date in enumerate(self.datetime):
            wdir_input = list()
            for j, angle in enumerate(np.linspace(-np.pi/4, 9*np.pi/4, num=6)):
                wdir_input.append(
                    np.asarray(
                        501 * [np.pi / 2 + angle + self.wind_deviation_angle])
                    )
            wdir_input = np.array(wdir_input).flatten()
            wdirfield = griddata(
                (self.r_input, self.theta_input),
                wdir_input,
                (r_out.flatten(), theta_out.flatten()),
                method='linear')
            # wdirfield interp
            idx = np.where(np.isnan(wdirfield))
            if len(idx[0]) > 0:
                _idx = np.where(~np.isnan(wdir_input))
                _wdirfield = griddata(
                    (self.r_input[_idx], self.theta_input[_idx]),
                    wdir_input[_idx],
                    (r_out.flatten()[idx], theta_out.flatten()[idx]),
                    method='nearest')
                for i, _idx in enumerate(idx):
                    wdirfield[_idx] = _wdirfield[i]
            wdirfield = wdirfield.reshape(r_out.shape)
            fields[date] = wdirfield
        return fields

    @property
    @lru_cache
    def r_input(self):
        r_input = []
        for j, angle in enumerate(np.linspace(-np.pi/4, 9*np.pi/4, num=6)):
            r_input.append(np.linspace(0, 500000, 501))
        return np.array(r_input).flatten()

    @property
    @lru_cache
    def theta_input(self):
        theta_input = []
        for j, angle in enumerate(np.linspace(-np.pi/4, 9*np.pi/4, num=6)):
            theta_input.append(np.asarray(501 * [angle]))
        return np.array(theta_input).flatten()

    @property
    @lru_cache
    def lon_arrays(self):
        lon_arrays = []
        for date, field in self.lonlatfields.items():
            lon_arrays.append(field['lon'])
        return lon_arrays

    @property
    @lru_cache
    def lat_arrays(self):
        lat_arrays = []
        for date, field in self.lonlatfields.items():
            lat_arrays.append(field['lat'])
        return lat_arrays

    @property
    @lru_cache
    def u_arrays(self):
        u_arrays = []
        for date in self.datetime:
            u_arrays.append(self.wfields[date] * np.cos(self.wdirfields[date]))
        return u_arrays

    @property
    @lru_cache
    def v_arrays(self):
        v_arrays = []
        for date in self.datetime:
            v_arrays.append(self.wfields[date] * np.sin(self.wdirfields[date]))
        return v_arrays

    @property
    def wind_deviation_angle(self):
        try:
            return self.__wind_deviation_angle
        except AttributeError:
            return np.pi / 9.

    @property
    @lru_cache
    def wind_directions(self):
        wind_directions = list()
        for date in self.datetime:
            wind_directions.append(self.wdirfields[date])
        return wind_directions

    @property
    def eye_lon(self):
        try:
            return self.__eye_lon
        except AttributeError:
            self._init_eye_data()
            return self.__eye_lon

    @property
    def eye_lat(self):
        try:
            return self.__eye_lat
        except AttributeError:
            self._init_eye_data()
            return self.__eye_lat

    @property
    def eye_x(self):
        try:
            return self.__eye_x
        except AttributeError:
            self._init_eye_data()
            return self.__eye_x

    @property
    def eye_y(self):
        try:
            return self.__eye_y
        except AttributeError:
            self._init_eye_data()
            return self.__eye_y

    @property
    @lru_cache
    def lonlatfields(self):
        # initialize coordinate transformation objects
        Mercator_to_WGS84 = Transformer.from_proj(
            'EPSG:3395',
            'EPSG:4326',
            always_xy=True,
            )
        # precompute output base meshgrid in cartesian coordinates
        x, y = pol2cart(*np.meshgrid(
                np.linspace(0, 500000, 501),
                np.linspace(0, 2.*np.pi, 501)
                ))
        # add eye location to the base grid at each datetime
        fields = {}
        for i, date in enumerate(self.datetime):
            lon, lat = Mercator_to_WGS84.transform(
                self.eye_x[i] + x,
                self.eye_y[i] + y)
            fields[date] = {'lon': lon, 'lat': lat}
        return fields

    @wind_deviation_angle.setter
    def wind_deviation_angle(self, wind_deviation_angle):
        self.__wind_deviation_angle = float(wind_deviation_angle)

    def _init_eye_data(self):
        eye_lon = []
        eye_lat = []
        for i, date in enumerate(self.datetime):
            eye_lon.append(self.tcprof[self.event_id]['param'][i][9, 1])
            eye_lat.append(self.tcprof[self.event_id]['param'][i][9, 0])
        # precompute eye location in mercator coordinates
        WGS84_to_Mercator = Transformer.from_proj(
            'EPSG:4326',
            'EPSG:3395',
            always_xy=True,
            )
        eye_x, eye_y = WGS84_to_Mercator.transform(eye_lon, eye_lat)
        self.__eye_x = eye_x
        self.__eye_y = eye_y
        self.__eye_lon = eye_lon
        self.__eye_lat = eye_lat

    @property
    def _event_id(self):
        return self.__event_id

    @_event_id.setter
    def _event_id(self, event_id):
        msg = f'event_id must be one of {list(self.tcprof.keys())}'
        assert event_id.capitalize() in self.tcprof, msg
        self.__event_id = event_id

    def _interp_hgrid(self, hgrid, timestep, nskip=1):

        # get list of steps
        wtiminc = timedelta(seconds=timestep*nskip)
        total_time = self.datetime[-1] - self.datetime[0]
        current = timedelta(seconds=0)
        steps = []
        while current.total_seconds() < total_time.total_seconds():
            steps.append(current)
            current += wtiminc
        if steps[-1].total_seconds() < total_time.total_seconds():
            steps.append(current)

        # create "times" vector
        times = [(t - self.datetime[0]).total_seconds() for t in self.datetime]

        # create x, y interpolators
        x = interp1d(times, self.lon_arrays, axis=0)
        y = interp1d(times, self.lat_arrays, axis=0)

        # create pressure interpolator
        p = interp1d(times, self.pressure, axis=0)

        # create the w, d interpolator
        w = interp1d(times, self.wind_magnitude, axis=0)
        d = interp1d(times, self.wind_directions, axis=0)

        return steps, x, y, w, d, p


# auxilliary functions
def pol2cart(rho, phi):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return(x, y)


# interface functions
def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('event_id')
    parser.add_argument('--overwrite', action='store_true')

    # init subparsers
    subparsers = parser.add_subparsers(dest='mode')
    subparsers.required = True

    # data plotting subparsers
    plot_data = subparsers.add_parser('plot_data')
    plot_data.add_argument('plot_type', choices=['wind', 'press'])
    plot_data.add_argument('--index', type=int)

    # mesh plotting subparsers
    plot_mesh = subparsers.add_parser('plot_mesh')
    plot_mesh.add_argument('plot_type', choices=['wind', 'press'])
    plot_mesh.add_argument('--index', type=int)

    # anim subparsers
    save_anim = subparsers.add_parser('save_anim')
    save_anim.add_argument('plot_type', choices=['wind', 'press'])

    # interp subparsers
    interp = subparsers.add_parser('interp')
    interp.add_argument('timestep', type=float)
    interp.add_argument('hgrid')
    interp.add_argument('crs')
    interp.add_argument('output_path',)
    interp.add_argument('--nskip', type=int, default=1)

    # parse args
    args = parser.parse_args()

    return args


def interp(args, tcprof):
    if args.output_path is None:
        output_path = (pathlib.Path('.') / 'wind.th').resolve()
    else:
        output_path = pathlib.Path(args.output_path).resolve()
    tcprof.interpolate(
        Hgrid.open(args.hgrid, crs=args.crs),
        output_path,
        args.timestep,
        nskip=args.nskip,
        overwrite=args.overwrite
        )


def plot(args, tcprof):
    if args.plot_type == 'wind':
        if args.index is not None:
            tcprof.plot_wfield(args.index)
        else:
            msg = "Need to implement wind movie."
            raise NotImplementedError(msg)
    elif args.plot_type == 'press':
        if args.index is not None:
            tcprof.plot_pfield(args.index)
        else:
            msg = "Need to implement pressure movie."
            raise NotImplementedError(msg)


def save_anim(args, tcprof):
    if args.output_path is None:
        output_path = (
            pathlib.Path('.') / f'{args.anim_type}.gif').resolve()
    else:
        output_path = pathlib.Path(output_path).resolve()

    if args.anim_type == 'wind':
        tcprof.make_anim_wfield(args.output_directory)


def mesh_plot(args, tcprof):
    msg = 'Need to implement mesh plotting'
    raise NotImplementedError(msg)


def main():
    args = parse_args()
    tcprof = TCProf(args.event_id)

    # mesh interpolation mode
    if args.mode == 'interp':
        interp(args, tcprof)

    # tcprof data plot mode
    elif args.mode == 'plot_data':
        plot(args, tcprof)

    # tcprof gif generation mode
    elif args.mode == 'save_anim':
        save_anim(args, tcprof)

    # result visualization mode
    elif args.mode == 'plot_mesh':
        mesh_plot(args, tcprof)

    # duck-type error (should be unreachable)
    else:
        msg = f"Unrecognized mode: {args.mode}."
        raise Exception(msg)

    return 0


def init():
    if __name__ == '__main__':
        try:
            import colored_traceback
            colored_traceback.add_hook(always=True)
        except ModuleNotFoundError:
            pass
        exit(main())


init()



from datetime import datetime, timezone, timedelta
import pathlib
from typing import Union, List
import warnings

import f90nml
import numpy as np


class StationsOutput:

    def __init__(self, outputs: Union[str, pathlib.Path],
                 stations_file: Union[str, pathlib.Path] = None,
                 ):
        outputs = pathlib.Path(outputs)
        if not outputs.is_dir():
            raise Exception(f'{str(outputs)} is not a directory.')
        self.__station_id: List = []
        if stations_file is None:
            stations_file = outputs.resolve() / '../station.in'
            if not stations_file.is_file():
                stations_file = None
            else:
                with open(stations_file) as f:
                    f.readline()
                    for station in range(int(f.readline())):
                        line = f.readline()
                        if '!' in line:
                            self.__station_id.append(line.split('!')[-1])
                        else:
                            self.__station_id.append(None)
        self._manifest = list(outputs.glob('staout_*'))

        # find the start_date. Let's first check if it's on the param.nml file
        param_file = outputs.resolve() / '../param.nml'
        if param_file.is_file():
            param = f90nml.read(param_file)
            start_year = param['opt'].get('start_year')
            start_month = param['opt'].get('start_month')
            start_day = param['opt'].get('start_day')
            start_hour = param['opt'].get('start_hour')
            utc_start = param['opt'].get('utc_start')
            if None in [start_year, start_month, start_day, start_hour,
                        utc_start]:
                warnings.warn('Could not determine start date automatically.')
                self.__start_date = None
            else:
                self.__start_date = datetime(start_year, start_month,
                                             start_day, int(start_hour), 0,
                                             tzinfo=timezone(timedelta(
                                                hours=-utc_start)))
        self.__rndays = None
        for file in self._manifest:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                data = np.loadtxt(file)
                if len(w) != 0:
                    pass
                else:
                    self.__rndays = timedelta(seconds=data[-1, 0])
                    break

        if self.rndays is None:
            raise Exception("Ouptut directory doesn't contain any station "
                            "output data.")

    def set_start_date(self, start_date: datetime):
        self.__start_date = start_date

    def get_station_id(self, index):
        return self.__station_id[index].strip()

    @property
    def start_date(self):
        return self.__start_date

    @property
    def rndays(self):
        return self.__rndays



from enum import Enum


class BcType(Enum):
    ELEVATION = 'iettype'
    FLOW = 'ifltype'
    TEMPERATURE = 'itetype'
    SALINITY = 'isatype'
    TRACER = 'itrtype'


class InitialElevationType(Enum):
    NONE = 0
    TIME_VARYING = 1
    CONSTANT = 2
    TIDAL = 3
    SPACE_TIME_VARYING = 4
    TIDAL_AND_SPACE_TIME_VARYING = 5


class InitialFlowType(Enum):
    NONE = 0
    TIME_VARYING = 1
    CONSTANT = 2
    TIDAL = 3
    SPACE_TIME_VARYING = 4
    SPACE_TIME_VARYING_3D = -4
    TIDAL_AND_SPACE_TIME_VARYING_3D = 5
    FLANTHER = -1


class InitialTemperatureType(Enum):
    NONE = 0
    TIME_VARYING = 1
    CONSTANT = 2
    INITIAL_PROFILE_FOR_INFLOW = 3
    INPUT_3D = 4


class InitialSalinityType(Enum):
    NONE = 0
    TIME_VARYING = 1
    CONSTANT = 2
    INITIAL_PROFILE_FOR_INFLOW = 3
    INPUT_3D = 4


class InitialTracerType(Enum):
    NONE = 0
    TIME_VARYING = 1
    CONSTANT = 2
    INITIAL_PROFILE_FOR_INFLOW = 3
    INPUT_3D = 4


class InitialConditionType(Enum):
    ELEVATION = InitialElevationType
    FLOW = InitialFlowType
    TEMPERATURE = InitialTemperatureType
    SALINITY = InitialSalinityType
    TRACER = InitialTracerType


class BoundaryCondition:

    def __init__(
            self,
            iettype: InitialElevationType = InitialElevationType.NONE,
            ifltype: InitialFlowType = InitialFlowType.NONE,
            itetype: InitialTemperatureType = InitialTemperatureType.NONE,
            isatype: InitialSalinityType = InitialSalinityType.NONE,
            itrtype: InitialTracerType = InitialTracerType.NONE
    ):
        self.iettype = iettype
        self.ifltype = ifltype
        self.itetype = itetype
        self.isatype = isatype
        self.itrtype = itrtype

    def __str__(self):
        return f"{self.iettype.value} " \
               f"{self.ifltype.value} " \
               f"{self.itetype.value} " \
               f"{self.isatype.value} " \
               f"{self.itrtype.value}"



from abc import ABC
from collections import defaultdict
from functools import lru_cache
import logging
from itertools import permutations
import os
import pathlib
from typing import Union, Sequence, Hashable, List, Dict

import geopandas as gpd  # type: ignore[import]
from matplotlib.collections import PolyCollection  # type: ignore[import]
from matplotlib.path import Path  # type: ignore[import]
from matplotlib.tri import Triangulation  # type: ignore[import]
from matplotlib.transforms import Bbox  # type: ignore[import]
import numpy as np  # type: ignore[import]
from pyproj import Transformer, CRS  # type: ignore[import]
from shapely.geometry import (  # type: ignore[import]
    MultiPolygon,
    Polygon,
    LineString,
    LinearRing,
)

from pyschism.mesh.parsers import grd
from pyschism.figures import figure

_logger = logging.getLogger(__name__)

class Description:

    def __set__(self, obj, description):
        if description is None:
            description = ''
        if not isinstance(description, str):
            raise TypeError('Argument description must be a string, not '
                            f'type {type(description)}.')
        obj.__dict__['description'] = description.replace('\n', ' ')

    def __get__(self, obj, val):
        return obj.__dict__['description']


class Nodes:

    def __set__(self, obj: "Gr3", nodes: Dict[Hashable, List[List]]):
        """Setter for the nodes attribute.

        Argument nodes must be of the form:
            {id: [(x0, y0), z0]}
            or
            {id: [(x0, y0), [z0, ..., zn]}

        Gr3 format is assumed to be exclusively a 2D format that can hold
        triangles or quads.

        """
        for coords, _ in nodes.values():
            if len(coords) != 2:
                raise ValueError(
                    'Coordinate vertices for a gr3 type must be 2D, but got '
                    f'coordinates {coords}.')
        obj.__dict__['nodes'] = nodes
        self.gr3 = obj
        # self.ball = NodeBall(self.gr3)

        obj.__dict__['id_to_index'] = {
            self.id()[i]: i for i in range(len(self.id()))}

        obj.__dict__['index_to_id'] = {
            i: self.id()[i] for i in range(len(self.id()))}

    def __call__(self):
        return self.gr3.__dict__['nodes']

    def id(self):
        node_id = self.gr3.__dict__.get('node_id')
        if node_id is None:
            node_id = list(self().keys())
            self.gr3.__dict__['node_id'] = node_id
        return node_id

    def index(self):
        node_index = self.gr3.__dict__.get('node_index')
        if node_index is None:
            node_index = np.arange(len(self()))
            self.gr3.__dict__['node_index'] = node_index
        return node_index

    def coord(self):
        coord = self.gr3.__dict__.get('coord')
        if coord is None:
            coord = np.array([coords for coords, _ in self().values()])
            self.gr3.__dict__['coord'] = coord
        return coord

    def values(self):
        values = self.gr3.__dict__.get('values')
        if values is None:
            values = np.array([val for _, val in self().values()])
            self.gr3.__dict__['values'] = values
        return values

    def get_index_by_id(self, id: Hashable):
        return self.gr3.__dict__['id_to_index'][id]

    def get_id_by_index(self, index: int):
        return self.gr3.__dict__['index_to_id'][index]

    def get_indexes_around_index(self, index):
        indexes_around_index = self.__dict__.get('indexes_around_index')
        if indexes_around_index is None:
            def append(geom):
                for simplex in geom:
                    for i, j in permutations(simplex, 2):
                        indexes_around_index[i].add(j)
            indexes_around_index = defaultdict(set)
            append(self.gr3.elements.triangles())
            append(self.gr3.elements.quads())
            self.__dict__['indexes_around_index'] = indexes_around_index
        return list(indexes_around_index[index])


class Elements:

    def __set__(self, obj: "Gr3", elements: Dict[Hashable, Sequence]):
        if not isinstance(elements, dict):
            raise TypeError('Argument elements must be a dict.')
        vertex_id_set = set(obj.nodes.id())
        for i, element in enumerate(elements):
            if not isinstance(element, Sequence):
                raise TypeError(f'Element with index {i} of the elements '
                                f'argument must be of type {Sequence}, not '
                                f'type {type(element)}.')
            if not set(element).issubset(vertex_id_set):
                ValueError(f'Element with index {i} is not a subset of the '
                           "coordinate id's.")
        obj.__dict__['elements'] = elements
        self.gr3 = obj

    def __call__(self):
        return self.gr3.__dict__["elements"]

    def id(self):
        element_id = self.gr3.__dict__.get('element_id')
        if element_id is None:
            element_id = list(self().keys())
            self.gr3.__dict__['element_id'] = element_id
        return element_id

    def index(self):
        element_index = self.gr3.__dict__.get('element_index')
        if element_index is None:
            element_index = np.arange(len(self()))
            self.gr3.__dict__['element_index'] = element_index
        return element_index

    def array(self):
        element_array = self.gr3.__dict__.get('element_array')
        if element_array is None:
            rank = int(max(map(len, self().values())))
            element_array = np.full((len(self()), rank), -1)
            for i, element in enumerate(self().values()):
                row = np.array(list(map(self.gr3.nodes.get_index_by_id, element)))
                element_array[i, :len(row)] = row
                element_array = np.ma.masked_equal(element_array, -1)
                self.gr3.__dict__['element_array'] = element_array
        return element_array

    def triangles(self):
        triangles = self.gr3.__dict__.get('triangles')
        if triangles is None:
            triangles = np.array(
                [list(map(self.gr3.nodes.get_index_by_id, element))
                for element in self().values()
                if len(element) == 3])
            self.gr3.__dict__['triangles'] = triangles
        return triangles

    def quads(self):
        quads = self.gr3.__dict__.get('quads')
        if quads is None:
            quads = np.array(
                [list(map(self.gr3.nodes.get_index_by_id, element))
                for element in self().values()
                if len(element) == 4])
            self.gr3.__dict__['quads'] = quads
        return quads

    def triangulation(self):
        triangles = self.triangles().tolist()
        for quad in self.quads():
            triangles.append([quad[0], quad[1], quad[3]])
            triangles.append([quad[1], quad[2], quad[3]])
        return Triangulation(
            self.gr3.nodes.coord()[:, 0],
            self.gr3.nodes.coord()[:, 1],
            triangles)

    def geodataframe(self):
        data = []
        for id, element in self().items():
            data.append({
                'geometry': Polygon(
                    self.gr3.nodes.coord()[list(
                        map(self.gr3.nodes.get_index_by_id, element))]),
                'id': id})
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)


class Crs:

    def __set__(self, obj, val):
        if val is not None:
            obj.__dict__['crs'] = CRS.from_user_input(val)

    def __get__(self, obj, val):
        return obj.__dict__.get('crs')


class Edges:

    def __init__(self, grd: "Gr3"):
        self.gr3 = grd

    @lru_cache(maxsize=1)
    def __call__(self) -> gpd.GeoDataFrame:
        data = []
        for ring in self.gr3.hull.rings().itertuples():
            coords = ring.geometry.coords
            for i in range(1, len(coords)):
                data.append({
                    "geometry": LineString([coords[i-1], coords[i]]),
                    "bnd_id": ring.bnd_id,
                    "type": ring.type})
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    def exterior(self):
        return self().loc[self()['type'] == 'exterior']

    def interior(self):
        return self().loc[self()['type'] == 'interior']


class Rings:

    def __init__(self, grd: "Gr3"):
        self.gr3 = grd

    @lru_cache(maxsize=1)
    def __call__(self) -> gpd.GeoDataFrame:
        tri = self.gr3.elements.triangulation()
        idxs = np.vstack(list(np.where(tri.neighbors == -1))).T
        boundary_edges = []
        for i, j in idxs:
            boundary_edges.append(
                (tri.triangles[i, j], tri.triangles[i, (j+1) % 3]))
        sorted_rings = sort_rings(edges_to_rings(boundary_edges),
                                  self.gr3.nodes.coord())
        data = []
        for bnd_id, rings in sorted_rings.items():
            coords = self.gr3.nodes.coord()[rings['exterior'][:, 0], :]
            geometry = LinearRing(coords)
            data.append({
                    "geometry": geometry,
                    "bnd_id": bnd_id,
                    "type": 'exterior'
                })
            for interior in rings['interiors']:
                coords = self.gr3.nodes.coord()[interior[:, 0], :]
                geometry = LinearRing(coords)
                data.append({
                    "geometry": geometry,
                    "bnd_id": bnd_id,
                    "type": 'interior'
                })
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    def exterior(self):
        return self().loc[self()['type'] == 'exterior']

    def interior(self):
        return self().loc[self()['type'] == 'interior']


class Hull:

    def __init__(self, grd: "Gr3"):
        self.gr3 = grd
        self.edges = Edges(grd)
        self.rings = Rings(grd)

    @lru_cache(maxsize=1)
    def __call__(self) -> gpd.GeoDataFrame:
        data = []
        for bnd_id in np.unique(self.rings()['bnd_id'].tolist()):
            exterior = self.rings().loc[
                (self.rings()['bnd_id'] == bnd_id) &
                (self.rings()['type'] == 'exterior')]
            interiors = self.rings().loc[
                (self.rings()['bnd_id'] == bnd_id) &
                (self.rings()['type'] == 'interior')]
            data.append({
                    "geometry": Polygon(
                        exterior.iloc[0].geometry.coords,
                        [row.geometry.coords for _, row
                            in interiors.iterrows()]),
                    "bnd_id": bnd_id
                })
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    @lru_cache(maxsize=1)
    def exterior(self):
        data = []
        for exterior in self.rings().loc[
                self.rings()['type'] == 'exterior'].itertuples():
            data.append({"geometry": Polygon(exterior.geometry.coords)})
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    @lru_cache(maxsize=1)
    def interior(self):
        data = []
        for interior in self.rings().loc[
                self.rings()['type'] == 'interior'].itertuples():
            data.append({"geometry": Polygon(interior.geometry.coords)})
        return gpd.GeoDataFrame(data, crs=self.gr3.crs)

    @lru_cache(maxsize=1)
    def implode(self) -> gpd.GeoDataFrame:
        return gpd.GeoDataFrame(
            {"geometry": MultiPolygon([polygon.geometry for polygon
                                       in self().itertuples()])},
            crs=self.gr3.crs)

    def multipolygon(self) -> MultiPolygon:
        return self.implode().iloc[0].geometry


class Gr3(ABC):

    _description = Description()
    _nodes = Nodes()
    _elements = Elements()
    _crs = Crs()

    def __init__(self, nodes, elements=None, description=None, crs=None):
        self._nodes = nodes
        self._elements = elements
        self._description = description
        self._crs = crs
        self._hull = Hull(self)

    def __str__(self):
        return grd.to_string(**self.to_dict())

    def __repr__(self):
        return ", ".join([
            f"{self.nodes()}",
            f"elements={self.elements()}",
            f"description={self.description}",
            f"crs={self.crs}"])

    def to_dict(self):
        return {
            "description": self.description,
            "nodes": self.nodes(),
            "elements": self.elements(),
            "crs": self.crs}

    def write(self, path, overwrite=False):
        grd.write(self.to_dict(), path, overwrite)

    def get_x(self, crs: Union[CRS, str] = None):
        return self.get_xy(crs)[:, 0]

    def get_y(self, crs: Union[CRS, str] = None):
        return self.get_xy(crs)[:, 1]

    def get_xy(self, crs: Union[CRS, str] = None):
        if crs is not None:
            crs = CRS.from_user_input(crs)
            if not crs.equals(self.crs):
                transformer = Transformer.from_crs(
                    self.crs, crs, always_xy=True)
                x, y = transformer.transform(self.x, self.y)
                return np.vstack([x, y]).T
        return np.vstack([self.x, self.y]).T

    def get_bbox(self, crs: Union[CRS, str] = None) -> Bbox:
        vertices = self.get_xy(crs)
        x0 = np.min(vertices[:, 0])
        x1 = np.max(vertices[:, 0])
        y0 = np.min(vertices[:, 1])
        y1 = np.max(vertices[:, 1])
        return Bbox([[x0, y0], [x1, y1]])

    def transform_to(self, dst_crs):
        """Transforms coordinate system of mesh in-place.
        """
        dst_crs = CRS.from_user_input(dst_crs)
        if not self.crs.equals(dst_crs):
            xy = self.get_xy(dst_crs)
            self._nodes = {self.nodes.id()[i]:
                           (coord.tolist(), self.nodes.values()[i])
                           for i, coord in enumerate(xy)}
            self._crs = dst_crs

    def vertices_around_vertex(self, index):
        return self.nodes.vertices_around_vertex(index)

    @classmethod
    def open(cls, file: Union[str, os.PathLike],
             crs: Union[str, CRS] = None):
        return cls(**grd.read(pathlib.Path(file), boundaries=False))

    @figure
    def tricontourf(self, axes=None, show=True, figsize=None, **kwargs):
        if len(self.triangles) > 0:
            axes.tricontourf(self.triangulation, self.values, **kwargs)
        return axes

    @figure
    def tripcolor(self, axes=None, show=True, figsize=None, **kwargs):
        if len(self.triangles) > 0:
            axes.tripcolor(self.triangulation, self.values, **kwargs)
        return axes

    @figure
    def triplot(
        self,
        axes=None,
        show=False,
        figsize=None,
        linewidth=0.07,
        color='black',
        **kwargs
    ):
        if len(self.triangles) > 0:
            kwargs.update({'linewidth': linewidth})
            kwargs.update({'color': color})
            axes.triplot(self.triangulation, **kwargs)
        return axes

    @figure
    def quadplot(
        self,
        axes=None,
        show=False,
        figsize=None,
        facecolor='none',
        edgecolor='k',
        linewidth=0.07,
        **kwargs
    ):
        if len(self.quads) > 0:
            pc = PolyCollection(
                self.coords[self.quads],
                facecolor=facecolor,
                edgecolor=edgecolor,
                linewidth=0.07,
            )
            axes.add_collection(pc)
        return axes

    @figure
    def quadface(
        self,
        axes=None,
        show=False,
        figsize=None,
        **kwargs
    ):
        if len(self.quads) > 0:
            pc = PolyCollection(
                self.coords[self.quads],
                **kwargs
            )
            quad_value = np.mean(self.values[self.quads], axis=1)
            pc.set_array(quad_value)
            axes.add_collection(pc)
        return axes

    @figure
    def plot_wireframe(self, axes=None, show=False, **kwargs):
        axes = self.triplot(axes=axes, **kwargs)
        axes = self.quadplot(axes=axes, **kwargs)
        return axes

    @property
    def nodes(self):
        return self._nodes

    @property
    def coords(self):
        return self._nodes.coord()

    @property
    def vertices(self):
        return self._nodes.coord()

    @property
    def vertex_id(self):
        return self._nodes.id()

    @property
    def elements(self):
        return self._elements

    @property
    def element_id(self):
        return self._elements.id()

    @property
    def values(self):
        return self._nodes.values()

    @property
    def description(self):
        return self._description

    @property
    def crs(self):
        return self._crs

    @property
    def x(self):
        return self._nodes.coord()[:, 0]

    @property
    def y(self):
        return self._nodes.coord()[:, 1]

    @property
    def hull(self):
        return self._hull

    @property
    def triangles(self):
        return self._elements.triangles()

    @property
    def quads(self):
        return self._elements.quads()

    @property
    def triangulation(self):
        return self._elements.triangulation()

    @property
    def bbox(self):
        return self.get_bbox()


def edges_to_rings(edges):
    if len(edges) == 0:
        return edges
    # start ordering the edges into linestrings
    edge_collection = list()
    ordered_edges = [edges.pop(-1)]
    e0, e1 = [list(t) for t in zip(*edges)]
    while len(edges) > 0:
        if ordered_edges[-1][1] in e0:
            idx = e0.index(ordered_edges[-1][1])
            ordered_edges.append(edges.pop(idx))
        elif ordered_edges[0][0] in e1:
            idx = e1.index(ordered_edges[0][0])
            ordered_edges.insert(0, edges.pop(idx))
        elif ordered_edges[-1][1] in e1:
            idx = e1.index(ordered_edges[-1][1])
            ordered_edges.append(
                list(reversed(edges.pop(idx))))
        elif ordered_edges[0][0] in e0:
            idx = e0.index(ordered_edges[0][0])
            ordered_edges.insert(
                0, list(reversed(edges.pop(idx))))
        else:
            edge_collection.append(tuple(ordered_edges))
            idx = -1
            ordered_edges = [edges.pop(idx)]
        e0.pop(idx)
        e1.pop(idx)
    # finalize
    if len(edge_collection) == 0 and len(edges) == 0:
        edge_collection.append(tuple(ordered_edges))
    else:
        edge_collection.append(tuple(ordered_edges))
    return edge_collection


def sort_rings(index_rings, vertices):
    """Sorts a list of index-rings.

    Takes a list of unsorted index rings and sorts them into an "exterior" and
    "interior" components. Any doubly-nested rings are considered exterior
    rings.

    TODO: Refactor and optimize. Calls that use :class:matplotlib.path.Path can
    probably be optimized using shapely.
    """

    # sort index_rings into corresponding "polygons"
    areas = list()
    for index_ring in index_rings:
        e0, e1 = [list(t) for t in zip(*index_ring)]
        areas.append(float(Polygon(vertices[e0, :]).area))

    # maximum area must be main mesh
    idx = areas.index(np.max(areas))
    exterior = index_rings.pop(idx)
    areas.pop(idx)
    _id = 0
    _index_rings = dict()
    _index_rings[_id] = {
        'exterior': np.asarray(exterior),
        'interiors': []
    }
    e0, e1 = [list(t) for t in zip(*exterior)]
    path = Path(vertices[e0 + [e0[0]], :], closed=True)
    while len(index_rings) > 0:
        # find all internal rings
        potential_interiors = list()
        for i, index_ring in enumerate(index_rings):
            e0, e1 = [list(t) for t in zip(*index_ring)]
            if path.contains_point(vertices[e0[0], :]):
                potential_interiors.append(i)
        # filter out nested rings
        real_interiors = list()
        for i, p_interior in reversed(
                list(enumerate(potential_interiors))):
            _p_interior = index_rings[p_interior]
            check = [index_rings[k]
                     for j, k in
                     reversed(list(enumerate(potential_interiors)))
                     if i != j]
            has_parent = False
            for _path in check:
                e0, e1 = [list(t) for t in zip(*_path)]
                _path = Path(vertices[e0 + [e0[0]], :], closed=True)
                if _path.contains_point(vertices[_p_interior[0][0], :]):
                    has_parent = True
            if not has_parent:
                real_interiors.append(p_interior)
        # pop real rings from collection
        for i in reversed(sorted(real_interiors)):
            _index_rings[_id]['interiors'].append(
                np.asarray(index_rings.pop(i)))
            areas.pop(i)
        # if no internal rings found, initialize next polygon
        if len(index_rings) > 0:
            idx = areas.index(np.max(areas))
            exterior = index_rings.pop(idx)
            areas.pop(idx)
            _id += 1
            _index_rings[_id] = {
                'exterior': np.asarray(exterior),
                'interiors': []
            }
            e0, e1 = [list(t) for t in zip(*exterior)]
            path = Path(vertices[e0 + [e0[0]], :], closed=True)
    return _index_rings


def signed_polygon_area(vertices):
    # https://code.activestate.com/recipes/578047-area-of-polygon-using-shoelace-formula/
    n = len(vertices)  # of vertices
    area = 0.0
    for i in range(n):
        j = (i + 1) % n
        area += vertices[i][0] * vertices[j][1]
        area -= vertices[j][0] * vertices[i][1]
        return area / 2.0



#! /usr/bin/env python
import unittest
import pathlib
import tempfile
from pyschism.mesh.sms2dm import reader, writer


class Sms2dmTestCase(unittest.TestCase):

    def setUp(self):
        self.nodes = {
            '1': ([0., 0.], -99999.),
            '2': ([.5, 0.], -99999.),
            '3': ([1., 0.], -99999.),
            '4': ([1., 1.], -99999.),
            '5': ([0., 1.], -99999.),
            '6': ([.5, 1.5], -99999.),
            '7': ([.33, .33], -99999.),
            '8': ([.66, .33], -99999.),
            '9': ([.5, .66], -99999.),
            '10': ([-1., 1.], -99999.),
            '11': ([-1., 0.], -99999.),
            }
        self.elements = {
            '1': ['5', '7', '9'],
            '2': ['1', '2', '7'],
            '3': ['2', '3', '8'],
            '4': ['8', '7', '2'],
            '5': ['3', '4', '8'],
            '6': ['4', '9', '8'],
            '7': ['4', '6', '5'],
            '8': ['5', '10', '11', '1'],
            '9': ['9', '4', '5'],
            '10': ['5', '1', '7']
            }

        self.boundaries = dict()

        self.boundaries[None] = {  # "open" boundaries
                0: {'indexes': ['10', '11', '1', '2']},
                1: {'indexes': ['2', '3', '4']}
        }

        self.boundaries[0] = {  # "land" boundaries
            0: {'indexes': ['4', '6']},
            1: {'indexes': ['6',  '5', '10']}
        }

        self.boundaries[1] = { # "interior" boundary
            0: {'indexes': ['7', '8', '9', '7']}
        }  


    def test_write_read(self):
        grd = {
            'ND': self.nodes,
            'E3T': {
                id: indexes 
                for id, indexes in self.elements.items() if len(indexes) == 3
            },
            'E4Q': {
                id: indexes 
                for id, indexes in self.elements.items() if len(indexes) == 4
            },
            'boundaries': self.boundaries,
        }
        tmpdir = tempfile.TemporaryDirectory()
        tmpfile = pathlib.Path(tmpdir.name) / 'hgrid.2dm'
        writer(grd, pathlib.Path(tmpfile))
        grd.pop('boundaries')
        self.assertDictEqual(reader(pathlib.Path(tmpfile)), grd)

    def test_write_raise_exists(self):
        grd = {
            'ND': self.nodes,
            'E3T': {
                id: indexes 
                for id, indexes in self.elements.items() if len(indexes) == 3
            },
            'E4Q': {
                id: indexes 
                for id, indexes in self.elements.items() if len(indexes) == 4
            }
        }
        tmpdir = tempfile.TemporaryDirectory()
        tmpfile = pathlib.Path(tmpdir.name) / 'hgrid.2dm'
        writer(grd, pathlib.Path(tmpfile))
        self.assertRaises(Exception, writer, grd, pathlib.Path(tmpfile))


if __name__ == '__main__':
    unittest.main()



from pyschism.forcing.atmosphere.nws.nws import NWS
from pyschism.forcing.atmosphere.nws import nws2
# from pyschism.forcing.atmosphere.nws.nws2.sflux import SfluxDownloadableDataset
__all__ = ["NWS", "nws2",
# "SfluxDownloadableDataset"
]



from abc import ABC, abstractmethod
# from datetime import datetime


class NWS(ABC):

    def __str__(self):
        """Returns string used in param.nml"""
        return f"{self.dtype.value}"

    def __call__(self, model_driver):
        self._start_date = model_driver.param.opt.start_date
        self._rnday = model_driver.param.core.rnday
        model_driver.param.opt.nws = self

    @abstractmethod
    def write(self, path, overwrite=False):
        """Provides a method for writting SCHISM atmospherics files to disk.

        Since the output is different for each NWS type, the derived class
        must implement this method.
        """

    @property
    @abstractmethod
    def dtype(self):
        """Returns the NWSType of this object.

        Derived classes must return the corresponding instance of
        :enum:`pyschism.enums.NWSType`
        """
